"""Taegis IPython Magics."""

import hashlib
import logging
import shlex
from argparse import ArgumentError, ArgumentParser
from pathlib import Path
from textwrap import dedent
from typing import Optional

import pandas as pd
from IPython.core.magic import Magics, line_cell_magic, line_magic, magics_class
from IPython.core.magic_arguments import argument, magic_arguments, parse_argstring
from IPython.display import display, display_markdown
from taegis_magic.cli import app
from taegis_magic.core.cache import (
    decode_base64_obj_as_pickle,
    display_cache,
    get_cache_item,
)
from taegis_magic.core.notebook import (
    find_notebook_name,
    generate_report,
    save_notebook,
)

from gql.transport.exceptions import TransportQueryError

log = logging.getLogger(__name__)


def taegis_magics_command_parser() -> ArgumentParser:
    """Magics support flags."""
    parser = ArgumentParser("taegis_magic_parser")
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        "--assign",
        metavar="NAME",
        type=str,
        help="Assign results as pandas DataFrame to NAME",
    )
    group.add_argument(
        "--append",
        metavar="NAME",
        type=str,
        help="Append results as pandas DataFrame to NAME",
    )
    parser.add_argument(
        "--display",
        metavar="NAME",
        help="Display NAME as markdown table",
    )
    parser.add_argument(
        "--cache",
        action="store_true",
        help="Save output to cache / Load output from cache (if present)",
    )

    return parser


@magics_class
class TaegisMagics(Magics):
    """Taegis Magics Class."""

    def __init__(self, shell=None, **kwargs):
        super().__init__(shell, **kwargs)

        log.debug("Trying to get notebook name for IPython shell...")
        # try to get the notebook name from the environment
        notebook_name = (
            self.shell.user_ns.get("TAEGIS_MAGIC_NOTEBOOK_FILENAME")
            or self.shell.user_ns.get("PAPERMILL_OUTPUT_PATH")
            or self.shell.user_ns.get("PAPERMILL_INPUT_PATH")
        )

        # try to find it through other means
        if not notebook_name:
            log.debug("Notebook name not found in environment variables...")
            try:
                notebook_name = find_notebook_name()
            except Exception as e:
                log.error(f"Error finding notebook name: {e}")
                notebook_name = None

        if notebook_name:
            log.debug(f"Notebook name found: {notebook_name}")
            self.shell.user_ns["TAEGIS_MAGIC_NOTEBOOK_FILENAME"] = notebook_name
            if not self.shell.user_ns.get("REPORT_TITLE"):
                self.shell.user_ns["REPORT_TITLE"] = Path(notebook_name).stem.title()
        else:
            log.error(
                "Could not determine notebook name.  Please set TAEGIS_MAGIC_NOTEBOOK_FILENAME manually."
            )

    @line_cell_magic
    def taegis(self, line: str, cell: Optional[str] = None):
        """Taegis Magics Line/Cell magic."""
        magic_args = None
        command_args = None
        notebook_filename = None

        args = shlex.split(line)
        parser = taegis_magics_command_parser()

        try:
            magic_args, command_args = parser.parse_known_args(args)
        except SystemExit:
            if "--help" in args or "-h" in args:
                command_args = args
            else:
                return
        except ArgumentError as e:
            print(e)
            return

        log.debug(f"Magic Args: {magic_args}")
        log.debug(f"Command Args: {command_args}")

        if magic_args and magic_args.cache:
            if not magic_args.assign:
                raise ValueError("--assign must be set with cache...")

            if (
                "TAEGIS_MAGIC_NOTEBOOK_FILENAME" in self.shell.user_ns
                and self.shell.user_ns["TAEGIS_MAGIC_NOTEBOOK_FILENAME"]
            ):
                notebook_filename = self.shell.user_ns["TAEGIS_MAGIC_NOTEBOOK_FILENAME"]
            else:
                notebook_filename = input("Notebook Filename:")

            if not notebook_filename:
                raise ValueError("Cannot determine file name of notebook...")

            notebook_fp = Path(notebook_filename)

            if notebook_fp.exists():
                self.shell.user_ns["TAEGIS_MAGIC_NOTEBOOK_FILENAME"] = notebook_filename
            else:
                raise ValueError(
                    "Notebook does not exist on disk, save notebook to disk before caching or "
                    "reload the taegis_magic extension (%reload_ext taegis_magic)..."
                )

            if not cell:
                cell = ""

            cache_digest = hashlib.sha256(bytes(line + cell, "utf-8")).hexdigest()
            cache = get_cache_item(notebook_fp, magic_args.assign, cache_digest)
            if cache:
                log.info(f"{magic_args.assign} found in cache...")
                log.debug("normalizing results...")
                data = decode_base64_obj_as_pickle(cache.get("data"))

                log.debug("converting to dataframe...")
                self.shell.user_ns[magic_args.assign] = pd.json_normalize(
                    data.results, max_level=3
                ).dropna(axis=1, how="all")

                log.info(f"re-setting {magic_args.assign}:{cache_digest} to cache...")
                display_cache(magic_args.assign, cache_digest, data)
                save_notebook()

                return

            log.info(f"{magic_args.assign} not found in cache...")

        if cell:
            command_args.extend(["--cell", cell])

        # os.environ["TAEGIS_MAGIC_OUTPUT"] = "True"
        try:
            result = app(command_args, prog_name="taegis", standalone_mode=False)
        except (SystemExit, TransportQueryError):
            result = None
        # os.environ["TAEGIS_MAGIC_OUTPUT"] = "False"

        if not result:
            return

        self.shell.user_ns["_taegis_magic_result"] = result

        if magic_args:
            if magic_args.assign:
                if isinstance(result.results, list):
                    self.shell.user_ns[magic_args.assign] = pd.json_normalize(
                        result.results, max_level=3
                    ).dropna(how="all", axis=1)
                else:
                    self.shell.user_ns[magic_args.assign] = result.results

            elif magic_args.append:
                if isinstance(result.results, list):
                    if magic_args.append not in self.shell.user_ns:
                        self.shell.user_ns[magic_args.append] = pd.DataFrame()

                    self.shell.user_ns[magic_args.append] = pd.concat(
                        [
                            self.shell.user_ns[magic_args.append],
                            pd.json_normalize(result.results, max_level=3).dropna(
                                how="all", axis=1
                            ),
                        ]
                    ).reset_index(drop=True)

                else:
                    if magic_args.append not in self.shell.user_ns:
                        self.shell.user_ns[magic_args.append] = []

                    self.shell.user_ns[magic_args.append].append(result.results)

            if magic_args.display:
                display_markdown(
                    self.shell.user_ns[magic_args.display].to_html(), raw=True
                )
                return

        if magic_args.cache:
            display_cache(magic_args.assign, cache_digest, result)
            save_notebook()
        else:
            display(result, exclude=["text/plain"])

    @magic_arguments()
    @argument(
        "--delay",
        type=int,
        default=0,
        help=("Delay in seconds while saving the notebook, defaults to 0 seconds"),
    )
    @line_magic
    def save_notebook(self, line: str):
        """Save the current notebook."""
        args = parse_argstring(self.save_notebook, line)
        save_notebook(delay=args.delay)

    @line_magic
    def generate_report(self, line: str = ""):
        """Save the current notebook as a report.

        Sets the TAEGIS_MAGIC_REPORT_FILENAME variable in the user namespace."""
        if (
            "TAEGIS_MAGIC_NOTEBOOK_FILENAME" not in self.shell.user_ns
            or not self.shell.user_ns["TAEGIS_MAGIC_NOTEBOOK_FILENAME"]
        ):
            raise ValueError(
                "Cannot determine file name of notebook. Please set TAEGIS_MAGIC_NOTEBOOK_FILENAME."
            )

        if not Path(self.shell.user_ns["TAEGIS_MAGIC_NOTEBOOK_FILENAME"]).exists():
            raise FileNotFoundError(
                dedent(
                    f"""Notebook {self.shell.user_ns['TAEGIS_MAGIC_NOTEBOOK_FILENAME']} does not exist.
                
                    Save notebook manually or run %save_notebook to save the notebook to disk.
                    """
                )
            )

        report = generate_report(
            filename=self.shell.user_ns["TAEGIS_MAGIC_NOTEBOOK_FILENAME"]
        )
        self.shell.user_ns["TAEGIS_MAGIC_REPORT_FILENAME"] = str(report.resolve())


def load_ipython_extension(ipython):
    """
    Load Taegis Magics extension.

    Parameters
    ----------
    ipython : InteractiveShell
        IPython session
    """
    ipython.register_magics(TaegisMagics)



"""Cache data within IPython output cells."""

import base64
import logging
from io import BytesIO
from pathlib import Path
from typing import Any, Dict, List, Tuple, Union

import compress_pickle
import nbformat
import pandas as pd
from IPython.display import display
from taegis_magic.core.normalizer import TaegisResultsNormalizer
from taegis_magic.commands.alerts import AlertsResultsNormalizer
from taegis_magic.commands.events import TaegisEventQueryNormalizer


logger = logging.getLogger(__name__)


def encode_obj_as_base64_pickle(obj: TaegisResultsNormalizer) -> str:
    """Encode any Taegis Normalized object as a base64 encoded bytes stream.

    Parameters
    ----------
    obj : TaegisResultsNormalizer
        Taegis Normalized Object

    Returns
    -------
    str
        base64 encoded bytes stream
    """
    with BytesIO() as bytes_io:
        compress_pickle.dump(obj, bytes_io, compression="lzma")
        return base64.b64encode(bytes_io.getvalue()).decode()


def decode_base64_obj_as_pickle(b64_string: str) -> TaegisResultsNormalizer:
    """Decode a base64 encoded bytes stream as an Taegis Normalized object.

    Parameters
    ----------
    b64_string : str
        base64 encoded bytes stream

    Returns
    -------
    TaegisResultsNormalizer
        Taegis Normalized object
    """
    return compress_pickle.loads(base64.b64decode(b64_string), compression="lzma")


def read_notebook(path: Union[str, Path]) -> nbformat.NotebookNode:
    """Parse `.ipynb` file into `NotebookNode` object.

    Parameters
    ----------
    path : Union[str, Path]
        Path to notebook file

    Returns
    -------
    nbformat.NotebookNode
        Notebook
    """
    if isinstance(path, str):
        path = Path(path).resolve()

    if not path.exists():
        raise EnvironmentError(f"{str(path)} does not exist...")

    return nbformat.reads(
        path.read_text(encoding="utf-8"), as_version=nbformat.current_nbformat
    )


def get_cache_list(path: Union[str, Path]) -> List[Tuple[str, str]]:
    """Get a list of cached object in a notebook.

    Parameters
    ----------
    path : Union[str, Path]
        Path to notebook

    Returns
    -------
    List[Tuple[str, str]]
        List of cached object names
    """
    nb = read_notebook(path)

    return [
        (
            (output.get("metadata", {}) or {}).get("name", ""),
            (output.get("metadata", {}) or {}).get("hash", ""),
        )
        for cell in (nb.cells or [])
        for output in (cell.get("outputs", []) or [])
    ]


def get_cache_item(
    path: Union[str, Path], name: str, cache_source_hash: str
) -> Dict[str, Any]:
    """Get named object from cache.

    Parameters
    ----------
    path : Union[str, Path]
        Path to notebook
    name : str
        Name of the cached object

    Returns
    -------
    Dict[str, Any]
        Cached object.
    """
    if name not in [item[0] for item in get_cache_list(path)]:
        logger.debug(f"{name} not found in {str(path)} cache...")
        return {}

    nb = read_notebook(path)

    try:
        cache = next(
            iter(
                [
                    (output.get("metadata", {}) or {})
                    for cell in (nb.cells or [])
                    for output in (cell.get("outputs", []) or [])
                    if output.get("metadata", {}).get("name") == name
                    and output.get("metadata", {}).get("hash") == cache_source_hash
                ]
            )
        )
    except Exception as e:
        logger.error(f"Error searching cache::{type(e).__name__}: {e}...")
        cache = {}

    return cache


def get_cached_objects(path: Union[str, Path]) -> List[TaegisResultsNormalizer]:
    """Get a list of all cached objects in a notebook.

    Parameters
    ----------
    path : Union[str, Path]
        Path to notebook

    Returns
    -------
    List[TaegisResultsNormalizer]
        List of Taegis Normalized objects
    """
    return [
        decode_base64_obj_as_pickle(
            get_cache_item(path=path, name=item[0], cache_source_hash=item[1]).get(
                "data", ""
            )
        )
        for item in get_cache_list(path)
        if item[0]
    ]


def notebook_contains_cached_results(path: Union[str, Path]) -> bool:
    """Check if a notebook contains all null queries.

    Parameters
    ----------
    path : Union[str, Path]
        Path to notebook

    Returns
    -------
    bool
        Does the notebook contain query results?

    Example
    -------
    Example::

        downloaded_notebooks = list(Path(DOWNLOAD_DIRECTORY).rglob('*.ipynb'))
        notebooks_with_null_findings = [
            notebook
            for notebook in downloaded_notebooks
            if notebook_is_null(notebook)
        ]
    """
    if (
        sum(
            [
                obj.results_returned
                for obj in get_cached_objects(path)
                if isinstance(
                    obj, (AlertsResultsNormalizer, TaegisEventQueryNormalizer)
                )
            ]
        )
        == 0
    ):
        return True

    return False


def display_cache(name: str, cache_digest: str, data: Any):
    """Display data and cache within output.

    Parameters
    ----------
    name : str
        Name of object.
    hash : str
        Unique hash for cache.
    data : Any
        Data to be cached.
    """
    display(
        data,
        metadata={
            "name": name,
            "data": encode_obj_as_base64_pickle(data),
            "hash": cache_digest,
        },
        exclude=["text/plain"],
    )



"""Taegis Magic preferences commands."""

import logging
from dataclasses import asdict, dataclass
from typing import Any, Dict, List, Optional

import typer
from dataclasses_json import dataclass_json
from taegis_magic.core.log import tracing
from taegis_magic.core.normalizer import TaegisResultsNormalizer

from taegis_magic.core.service import get_service

log = logging.getLogger(__name__)

app = typer.Typer(help="Taegis Preferences Commands.")


@dataclass_json
@dataclass
class PreferenceResultsNormalizer(TaegisResultsNormalizer):
    @property
    def results(self) -> List[Dict[str, Any]]:
        if not self.raw_results:
            return []

        if not isinstance(self.raw_results, list):
            return [asdict(self.raw_results)]

        return [asdict(preference) for preference in self.raw_results]


@app.command()
@tracing
def parent(region: Optional[str] = None):
    """
    Parent Preferences.

    Parameters
    ----------
    region : Optional[str], optional
        Taegis region, by default None

    Returns
    -------
    PreferenceResultsNormalizer
        PreferenceResultsNormalizer
    """
    service = get_service(environment=region)

    preferences = service.preferences.query.partner_preferences()

    results = PreferenceResultsNormalizer(
        service="preferences",
        tenant_id=service.tenant_id,
        region=service.environment,
        raw_results=preferences,
        arguments={"region": service.environment},
    )

    return results


@app.command()
@tracing
def ticketing_settings(region: Optional[str] = None):
    """
    Ticketing Settings Preferences.

    Parameters
    ----------
    region : Optional[str], optional
        Taegis region, by default None

    Returns
    -------
    PreferenceResultsNormalizer
        PreferenceResultsNormalizer
    """
    service = get_service(environment=region)

    settings = service.preferences.query.ticketing_settings()

    results = PreferenceResultsNormalizer(
        service="preferences",
        tenant_id=service.tenant_id,
        region=service.environment,
        raw_results=settings,
        arguments={"region": service.environment},
    )

    return results


@app.command()
@tracing
def user_preference(key: str, region: Optional[str] = None):
    """
    User Preferences by Key.

    Parameters
    ----------
    key : str
        Preference key
    region : Optional[str], optional
        Taegis region, by default None

    Returns
    -------
    PreferenceResultsNormalizer
        PreferenceResultsNormalizer
    """
    service = get_service(environment=region)

    preferences = service.preferences.query.user_preference_by_key(key=key)

    results = PreferenceResultsNormalizer(
        service="preferences",
        tenant_id=service.tenant_id,
        region=service.environment,
        raw_results=preferences,
        arguments={"key": key, "region": service.environment},
    )

    return results



import inspect
import logging
from typing import Optional

import typer
from typing_extensions import Annotated

from taegis_magic.core.log import tracing
from taegis_magic.core.normalizer import TaegisResult
from taegis_magic.core.service import get_service

app = typer.Typer(help="Taegis Subjects Commands.")

log = logging.getLogger(__name__)


@app.command()
@tracing
def current_subject(
    region: Annotated[Optional[str], typer.Option(help="Taegis region")] = None
):
    """Get current subject [TDRUser|Client] information."""
    service = get_service(environment=region)

    results = service.subjects.query.current_subject()

    normalized_results = TaegisResult(
        raw_results=results,
        service="users",
        tenant_id=service.tenant_id,
        region=service.environment,
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results



"""Pandas functions for Alerts DataFrames."""

import logging
import contextlib
import json
from typing import Dict, List, Optional

import pandas as pd
from taegis_magic.pandas.utils import coalesce_columns

log = logging.getLogger(__name__)


def convert_event_timestamps(
    df: pd.DataFrame, format_: str = "%Y-%m-%dT%H:%M:%SZ"
) -> pd.DataFrame:
    """Takes an Events dataframe and converts all the metadata time columns
    into a readable time format. All columns with usec or us will be dropped to reduce
    the amount of columns in the dataframe.

    Parameters
    ----------
    df : pd.DataFrame
        Events Dataframe
    format : str, optional
        Datetime string format to be used, by default "%Y-%m-%dT%H:%M:%SZ"

    Returns
    -------
    pd.DataFrame
        Returns Events Dataframe with added taegis_magic timestamp columns
    """

    if df.empty:
        return df

    df = df.copy()

    for column in [
        column
        for column in df.columns
        if column.endswith("_time_usec")
        or column.endswith("mod_time_us")
        and not column.startswith("taegis_magic.")
    ]:
        try:
            df[f"taegis_magic.{column}"] = pd.to_datetime(
                df[column], errors="ignore", unit="us"
            ).dt.strftime(format_)
        except Exception as exc:
            log.error(exc)
            continue

        df[f"taegis_magic.{column}"] = df[f"taegis_magic.{column}"].fillna("N/A")

    return df


def inflate_original_data(
    df: pd.DataFrame, original_data_columns: Optional[List[str]] = None
) -> pd.DataFrame:
    """Takes a DataFrame containing a valid Event original_data column
    and concats the nested JSON blob containing the
    event original data as columns. Returns a DataFrame.

    Parameters
    ----------
    df : pd.DataFrame
        Taegis Alerts or Events Query Dataframe
    original_data_columns : Optional[List[str]], optional
        A list of columns that contain Taegis Event original data, by default None

    Returns
    -------
    pd.DataFrame
        Dataframe that contains inflated Event original_data columns
    """

    def load_json(event):
        if isinstance(event, str):
            with contextlib.suppress(Exception):
                return json.loads(event)
        return {}

    if df.empty:
        return df

    if original_data_columns is None:
        original_data_columns = [
            "original_data",
            "event_data.original_data",
        ]

    valid_original_data_columns = []

    for original_data_col in original_data_columns:
        if original_data_col in df.columns:
            valid_original_data_columns.append(original_data_col)

    if not valid_original_data_columns:
        raise ValueError(
            f"DataFrame does not contain a vaild original data column: {original_data_columns}"
        )

    if any(df.columns.str.startswith("original_data.")) is False:
        return pd.concat(
            [
                df,
                pd.json_normalize(
                    coalesce_columns(df, valid_original_data_columns).apply(load_json)
                ).add_prefix("original_data."),
            ],
            axis=1,
        )

    return df


def inflate_schema_keys(df: pd.DataFrame) -> pd.DataFrame:
    """Parse and format event schema keys.

    Expected Format: 'scwx.schema.key'

    Parameters
    ----------
    df : pd.DataFrame
        Schema Keys

    Returns
    -------
    pd.DataFrame
        Schema Keys with formatted keys
    """
    if df.empty:
        return df

    if "key" not in df.columns:
        raise ValueError("DataFrame does not contain a 'key' column")

    df = df.copy()

    def split_key(key: str, index: int) -> str:
        if not isinstance(key, str):
            return "Error"

        parts = key.split(".", maxsplit=2)

        try:
            part = parts[index]
        except IndexError:
            return "Error"

        return part

    df["taegis_magic.schema"] = df["key"].apply(split_key, index=1)
    df["taegis_magic.key"] = df["key"].apply(split_key, index=2)

    return df



"""Pandas functions for investigations."""

import logging
from dataclasses import asdict

import pandas as pd
from taegis_magic.commands.utils.investigations import InvestigationEvidenceType
from taegis_magic.core.service import get_service
from taegis_magic.core.utils import to_dataframe
from taegis_magic.pandas.utils import chunk_list
from taegis_sdk_python.services.alerts.types import GetByIDRequestInput
from taegis_sdk_python.services.queries.types import QLQueriesInput

log = logging.getLogger(__name__)


def inflate_evidence(df: pd.DataFrame) -> pd.DataFrame:
    """Inflate evidence DataFrame by expanding the 'evidence' columns."""
    evidence_columns = {"alerts_evidence", "events_evidence", "search_queries_evidence"}
    if not set(df.columns).issuperset(evidence_columns):
        raise ValueError(
            f"DataFrame must contain the following columns: {evidence_columns}"
        )

    alerts_evidence = pd.json_normalize(df["alerts_evidence"].explode())
    if not alerts_evidence.empty:
        alerts_evidence["taegis_magic.evidence_type"] = InvestigationEvidenceType.Alert
        alerts_evidence["taegis_magic.evidence_id"] = alerts_evidence["alert_id"]

    events_evidence = pd.json_normalize(df["events_evidence"].explode())
    if not events_evidence.empty:
        events_evidence["taegis_magic.evidence_type"] = InvestigationEvidenceType.Event
        events_evidence["taegis_magic.evidence_id"] = events_evidence["event_id"]

    search_query_evidence = pd.json_normalize(df["search_queries_evidence"].explode())
    if not search_query_evidence.empty:
        search_query_evidence[
            "taegis_magic.evidence_type"
        ] = InvestigationEvidenceType.Query
        search_query_evidence["taegis_magic.evidence_id"] = search_query_evidence[
            "search_query"
        ]

    # Reset index to maintain DataFrame integrity
    evidence_df = pd.concat(
        [alerts_evidence, events_evidence, search_query_evidence], axis=0
    )
    evidence_df = evidence_df.dropna(axis=0, how="all").reset_index(drop=True)

    return evidence_df


def lookup_evidence(df: pd.DataFrame, region: str) -> pd.DataFrame:
    """Lookup evidence DataFrame by expanding the 'evidence' columns."""
    if (
        not "taegis_magic.evidence_type" in df.columns
        and not "taegis_magic.evidence_id" in df.columns
    ):
        raise ValueError(
            "DataFrame must contain the following columns: "
            "'taegis_magic.evidence_type', 'taegis_magic.evidence_id'"
        )

    service = get_service(environment=region)

    alerts_evidence = pd.DataFrame()
    events_evidence = pd.DataFrame()
    search_queries_evidence = pd.DataFrame()

    alerts_df = df[df["taegis_magic.evidence_type"] == InvestigationEvidenceType.Alert]
    events_df = df[df["taegis_magic.evidence_type"] == InvestigationEvidenceType.Event]
    queries_df = df[df["taegis_magic.evidence_type"] == InvestigationEvidenceType.Query]

    if not alerts_df.empty:
        for tenant_id in alerts_df["tenant_id"].unique():
            alerts_list = alerts_df[alerts_df["tenant_id"] == tenant_id][
                "taegis_magic.evidence_id"
            ].tolist()

            for chunk in chunk_list(alerts_list, 500):
                try:
                    with service(
                        tenant_id=tenant_id,
                        output="alerts { list { id tenant { id } status metadata { title severity confidence } entities { entities } } }",
                    ):
                        alerts = (
                            service.alerts.query.alerts_service_retrieve_alerts_by_id(
                                in_=GetByIDRequestInput(i_ds=chunk)
                            )
                        )

                    alerts_evidence = pd.concat(
                        [
                            alerts_evidence,
                            to_dataframe(
                                [asdict(alert) for alert in alerts.alerts.list]
                            ),
                        ]
                    )
                except Exception as e:
                    log.error(f"Error retrieving alerts for tenant {tenant_id}: {e}")
                    log.error(f"Alerts list: {chunk}")

        log.debug(f"Alerts evidence columns: {alerts_evidence.columns.tolist()}")
        log.debug(f"Alerts evidence retrieved: {alerts_evidence.head(5)}")

        df = df.merge(
            alerts_evidence,
            how="left",
            left_on="taegis_magic.evidence_id",
            right_on="id",
            suffixes=("", "_alerts"),
        )

    if not events_df.empty:
        for tenant_id in events_df["tenant_id"].unique():
            events_list = events_df[events_df["tenant_id"] == tenant_id][
                "taegis_magic.evidence_id"
            ].tolist()

            for chunk in chunk_list(events_list, 100):
                try:
                    with service(tenant_id=tenant_id):
                        events = service.events.query.events(ids=events_list)

                    events_evidence = pd.concat(
                        [
                            events_evidence,
                            to_dataframe([asdict(event) for event in events]),
                        ]
                    )
                except Exception as e:
                    log.error(f"Error retrieving events for tenant {tenant_id}: {e}")
                    log.error(f"Events list: {chunk}")

        log.debug(f"Events evidence columns: {events_evidence.columns.tolist()}")
        log.debug(f"Events evidence retrieved: {events_evidence.head(5)}")

        df = df.merge(
            events_evidence,
            how="left",
            left_on="taegis_magic.evidence_id",
            right_on="id",
            suffixes=("", "_events"),
        )

    if not queries_df.empty:
        for tenant_id in queries_df["tenant_id"].unique():
            queries_list = queries_df[queries_df["tenant_id"] == tenant_id][
                "taegis_magic.evidence_id"
            ].tolist()

            for chunk in chunk_list(queries_list, 100):
                try:
                    with service(tenant_id=tenant_id):
                        queries = service.queries.query.ql_queries(
                            input_=QLQueriesInput(rns=chunk)
                        )

                    search_queries_evidence = pd.concat(
                        [
                            search_queries_evidence,
                            to_dataframe([asdict(query) for query in queries.queries]),
                        ]
                    )
                except Exception as e:
                    log.error(f"Error retrieving queries for tenant {tenant_id}: {e}")
                    log.error(f"Queries list: {chunk}")

        log.debug(f"Query evidence columns: {search_queries_evidence.columns.tolist()}")
        log.debug(f"Query evidence retrieved: {search_queries_evidence.head(5)}")

        df = df.merge(
            search_queries_evidence,
            how="left",
            left_on="taegis_magic.evidence_id",
            right_on="rn",
            suffixes=("", "_queries"),
        )

    return df



import logging
import re
from concurrent.futures import ThreadPoolExecutor, as_completed
from dataclasses import asdict
from datetime import datetime, timedelta, timezone
from ipaddress import ip_address
from typing import Callable, List, Optional

import pandas as pd
from taegis_magic.core.utils import get_first_tenant_id
from taegis_magic.pandas.tenants import lookup_first_environment
from taegis_sdk_python import GraphQLService
from taegis_sdk_python.services.multi_tenant_ioc.types import (
    EventAggregationArguments,
    EventCountResult,
    LogicalType,
    LogicalTypeFilter,
    Operator,
    TenantsInput,
)
from taegis_magic.core.log import tracing

log = logging.getLogger(__name__)


@tracing
def normalize_event_count_results(input_: List[EventCountResult]) -> pd.DataFrame:
    """Normalize event count results into a DataFrame.

    Parameters
    ----------
    input_ : List[EventCountResult]

    Returns
    -------
    pd.DataFrame
    """
    df = pd.json_normalize(
        [
            asdict(result)
            for event_count_result in input_ or []
            for result in event_count_result.results or []
        ],
    )
    if "counts_by_tenant" in df.columns:
        df = df.explode("counts_by_tenant")
        df["counts_by_tenant.tenant_id"] = df["counts_by_tenant"].str.get("tenant_id")
        df["counts_by_tenant.count"] = df["counts_by_tenant"].str.get("count")
    else:
        log.warning('"No counts_by_tenant in results, skipping explode operation.')
    return df


@tracing
def is_domain(value: str) -> bool:
    """Check for a valid domain name.

    Parameters
    ----------
    value : str
        Value to validate.

    Returns
    -------
    bool
    """
    domain_regex = r"^(?!-)[A-Za-z0-9-]{1,63}(?<!-)\.[A-Za-z]{2,}$"
    return re.match(domain_regex, value) is not None


@tracing
def is_hash(value: str) -> bool:
    """Check for a valid hash value.

    Parameters
    ----------
    value : str
        Value to validate.

    Returns
    -------
    bool
    """

    def is_md5_hash(value: str) -> bool:
        return re.match(r"^[A-Fa-f0-9]{32}$", value) is not None

    def is_sha1_hash(value: str) -> bool:
        return re.match(r"^[A-Fa-f0-9]{40}$", value) is not None

    def is_sha256_hash(value: str) -> bool:
        return re.match(r"^[A-Fa-f0-9]{64}$", value) is not None

    return is_md5_hash(value) or is_sha1_hash(value) or is_sha256_hash(value)


@tracing
def is_ip_address(value: str) -> bool:
    """Check for valid IP address.

    Parameters
    ----------
    value : str
        Value to validate.

    Returns
    -------
    bool
    """
    try:
        ip_address(value)
        return True
    except ValueError:
        return False


@tracing
def event_count_by_logical_type_paginated(
    service: GraphQLService,
    *,
    arguments: EventAggregationArguments,
) -> List[EventCountResult]:
    """Get paginated event count by logical type.

    Parameters
    ----------
    service : GraphQLService
        The GraphQL service instance.
    arguments : EventAggregationArguments
        Arguments for the event count query.

    Returns
    -------
    List[EventCountResult]
        List of event count results.
    """
    pages = []
    results = service.multi_tenant_ioc.query.event_count_by_logical_type(
        arguments=arguments
    )
    pages.append(results)
    while results.next:
        results = service.multi_tenant_ioc.query.event_count_page(
            next_token=results.next
        )
        pages.append(results)

    return pages


@tracing
def search_ioc_domains(
    service: GraphQLService,
    *,
    iocs: List[str],
    days: int = 30,
    tenant_ids: Optional[List[str]] = None,
) -> Optional[List[EventCountResult]]:
    """Search for domain IOCs.

    Parameters
    ----------
    service : GraphQLService
    iocs : List[str]
        List of IOCs.
    days : int, optional
        Days back to search, by default 30
    tenant_ids : Optional[List[str]], optional
        Taegis Tenants IDs, by default None

    Returns
    -------
    Optional[List[EventCountResult]]
    """
    days += 1  # add 1 to days due to search failure on current day

    if not tenant_ids:
        tenant_ids = [get_first_tenant_id(service)]

    values = [ioc for ioc in iocs if is_domain(ioc)]

    if not values:
        log.error("No valid domains found in the provided IOCs.")
        return None

    arguments = EventAggregationArguments(
        logical_type_filter=LogicalTypeFilter(
            logical_type=LogicalType.DOMAIN, values=values, operator=Operator.EQUALS
        ),
        tenants_context=TenantsInput(tenant_ids=tenant_ids),
        earliest=(datetime.now(timezone.utc) - timedelta(days=days)).strftime("%Y%m%d"),
        latest=(datetime.now(timezone.utc) - timedelta(days=1)).strftime("%Y%m%d"),
    )
    results = event_count_by_logical_type_paginated(
        service=service,
        arguments=arguments,
    )

    return results


@tracing
def search_ioc_file_hashes(
    service: GraphQLService,
    *,
    iocs: List[str],
    days: int = 30,
    tenant_ids: Optional[List[str]] = None,
) -> Optional[EventCountResult]:
    """Search for File Hash IOCs.

    Parameters
    ----------
    service : GraphQLService
    iocs : List[str]
        List of IOCs.
    days : int, optional
        Days back to search, by default 30
    tenant_ids : Optional[List[str]], optional
        Taegis Tenants IDs, by default None

    Returns
    -------
    Optional[List[EventCountResult]]
    """
    days += 1  # add 1 to days due to search failure on current day

    if not tenant_ids:
        tenant_ids = [get_first_tenant_id(service)]

    values = [ioc for ioc in iocs if is_hash(ioc)]

    if not values:
        log.error("No valid hashes found in the provided IOCs.")
        return None

    arguments = EventAggregationArguments(
        logical_type_filter=LogicalTypeFilter(
            logical_type=LogicalType.HASH, values=values, operator=Operator.EQUALS
        ),
        tenants_context=TenantsInput(tenant_ids=tenant_ids),
        earliest=(datetime.now(timezone.utc) - timedelta(days=days)).strftime("%Y%m%d"),
        latest=(datetime.now(timezone.utc) - timedelta(days=1)).strftime("%Y%m%d"),
    )
    results = event_count_by_logical_type_paginated(
        service=service,
        arguments=arguments,
    )

    return results


@tracing
def search_ioc_ip_addresses(
    service: GraphQLService,
    *,
    iocs: List[str],
    days: int = 30,
    tenant_ids: Optional[List[str]] = None,
) -> Optional[EventCountResult]:
    """Search for IP Address IOCs.

    Parameters
    ----------
    service : GraphQLService
    iocs : List[str]
        List of IOCs.
    days : int, optional
        Days back to search, by default 30
    tenant_ids : Optional[List[str]], optional
        Taegis Tenants IDs, by default None

    Returns
    -------
    Optional[List[EventCountResult]]
    """
    days += 1  # add 1 to days due to search failure on current day

    if not tenant_ids:
        tenant_ids = [get_first_tenant_id(service)]

    values = [ioc for ioc in iocs if is_ip_address(ioc)]

    if not values:
        log.error("No valid IP addresses found in the provided IOCs.")
        return None

    arguments = EventAggregationArguments(
        logical_type_filter=LogicalTypeFilter(
            logical_type=LogicalType.IP, values=values, operator=Operator.EQUALS
        ),
        tenants_context=TenantsInput(tenant_ids=tenant_ids),
        earliest=(datetime.now(timezone.utc) - timedelta(days=days)).strftime("%Y%m%d"),
        latest=(datetime.now(timezone.utc) - timedelta(days=1)).strftime("%Y%m%d"),
    )
    results = event_count_by_logical_type_paginated(
        service=service,
        arguments=arguments,
    )

    return results


@tracing
def multi_tenant_ioc_search(
    service: GraphQLService,
    *,
    iocs: List[str],
    days: int = 30,
    tenant_ids: Optional[List[str]] = None,
    search_functions: Optional[Callable] = None,
) -> List[EventCountResult]:
    """Search for IOCs across multiple tenants.

    Parameters
    ----------
    service : GraphQLService
        The GraphQL service instance.
    iocs : List[str]
        List of IOCs to search for.
    days : int, optional
        Number of days to look back, by default 30.
    tenant_ids : Optional[List[str]], optional
        List of tenant IDs to search in, by default None.

    Returns
    -------
    Optional[List[EventCountResult]]
        List of event count results or None if no valid IOCs are found.
    """
    if not tenant_ids:
        tenant_ids = [get_first_tenant_id(service)]

    if not iocs:
        log.error("No IOCs provided for search.")
        return None

    results = []

    if not search_functions:
        search_functions = [
            search_ioc_domains,
            search_ioc_file_hashes,
            search_ioc_ip_addresses,
        ]

    for search_function in search_functions:
        if not callable(search_function):
            log.error(f"Search function {search_function} is not callable.")
            continue

        try:
            search_results = search_function(
                service=service, iocs=iocs, days=days, tenant_ids=tenant_ids
            )
            if search_results:
                results.extend(search_results)
        except Exception as e:
            log.error(f"Error during IOC search with {search_function.__name__}: {e}")

    return results


@tracing
def regioned_multi_tenant_ioc_search(
    service: GraphQLService,
    *,
    region: str,
    tenants: pd.DataFrame,
    iocs: List[str],
    days: int = 30,
    search_functions: Optional[Callable] = None,
) -> List[EventCountResult]:
    """Search for IOCs across multiple tenants by region.

    Parameters
    ----------
    service : GraphQLService
        The GraphQL service instance.
    region : str
        The Taegis region to search in.
    iocs : List[str]
        List of IOCs to search for.
    days : int, optional
        Number of days to look back, by default 30.
    tenant_ids : Optional[List[str]], optional
        List of tenant IDs to search in, by default None.

    Returns
    -------
    Optional[List[EventCountResult]]
        List of event count results or None if no valid IOCs are found.
    """
    if not "first_environment" in tenants.columns:
        log.error(
            "Tenants DataFrame must contain 'first_environment' column to filter by region."
        )
        return []

    with service(environment=region):
        tenant_ids = tenants[tenants["first_environment"] == region]["id"].tolist()

        return multi_tenant_ioc_search(
            service=service,
            iocs=iocs,
            days=days,
            tenant_ids=tenant_ids,
            search_functions=search_functions,
        )


@tracing
def threaded_multi_tenant_ioc_search(
    service: GraphQLService,
    *,
    tenants: pd.DataFrame,
    iocs: List[str],
    days: int = 30,
    search_functions: Optional[Callable] = None,
    **kwargs,
) -> pd.DataFrame:
    """Perform a multi-tenant IOC search using threading.

    Parameters
    ----------
    service : GraphQLService
        The GraphQL service instance.
    tenants : pd.DataFrame
        DataFrame containing tenant information.
    iocs : List[str]
        List of IOCs to search for.
    days : int, optional
        Number of days to look back, by default 30.
    search_functions : Optional[Callable], optional
        Custom search functions to use, by default None.

    Returns
    -------
    pd.DataFrame
        DataFrame containing the results of the IOC search.
    """
    if tenants.empty:
        log.error("No tenants provided for search.")
        return pd.DataFrame()

    if "first_environment" in tenants.columns:
        regions = tenants["first_environment"].unique().tolist()
    elif "environment" in tenants.columns:
        tenants["first_environment"] = tenants.pipe(lookup_first_environment)
        regions = tenants["first_environment"].unique().tolist()
    else:
        log.error(
            "Tenants DataFrame must contain 'first_environment' or 'environment' column."
        )
        return pd.DataFrame()

    if service.use_universal_authentication is False:
        for region in regions:
            with service(environment=region):
                service.access_token
    else:
        service.access_token

    future_results = {}

    with ThreadPoolExecutor(**kwargs) as executor:
        futures = {
            executor.submit(
                regioned_multi_tenant_ioc_search,
                service=service,
                region=region,
                tenants=tenants,
                iocs=iocs,
                days=days,
                search_functions=search_functions,
            ): region
            for region in regions
        }

        log.debug("Waiting for futures to complete...")
        for future in as_completed(futures):
            # stitch results per region
            region = futures[future]
            future_results[region] = future.result()
        log.debug("All futures completed.")

    log.debug("Normalizing results...")
    normalized_results = []
    for region, results in future_results.items():
        normalized_result = normalize_event_count_results(results)
        normalized_result["region"] = region
        normalized_results.append(normalized_result)
    log.debug("Results normalized.")

    if not normalized_results:
        log.error("No results found after normalization.")
        return pd.DataFrame()

    stiched_results = pd.concat(
        normalized_results,
        axis=1,
    ).reset_index(drop=True)

    return stiched_results



"""Taegis Tenants pandas functions."""

import logging
from dataclasses import asdict
from typing import Any, Dict, List, Optional

import pandas as pd
from taegis_magic.core.service import get_service
from taegis_magic.core.utils import get_tenant_id_column, to_dataframe
from taegis_sdk_python.services.tenants.types import TenantsQuery

log = logging.getLogger(__name__)


def inflate_environments(
    df: pd.DataFrame, columns: Optional[List[str]] = None
) -> pd.DataFrame:
    """Inflate environments to columns.

    Parameters
    ----------
    df : pd.DataFrame
        Tenants DateFrame

    Returns
    -------
    pd.DataFrame
        Inflated Tenants DateFrame

    Raises
    ------
    ValueError
        'environments' column not found in DataFrame
    """
    if df.empty:
        return df

    df = df.copy()

    if not columns:
        columns = ["tenant.environments", "environments"]

    environment_column = None
    for column in columns:
        if column in df.columns:
            environment_column = column
            break

    if not environment_column:
        raise ValueError(f"{columns} columns not found in DataFrame")

    def environments_to_dict(environments: List[Dict[str, Any]]) -> Dict[str, Any]:
        """Flatten environments array to dictionary key value pairs."""
        env_dict = {}
        for environment in environments:
            env_dict[environment.get("name", "error")] = environment.get(
                "enabled", False
            )
            env_dict[
                f"{environment.get('name', 'error')}.created_at"
            ] = environment.get("created_at", False)
            env_dict[
                f"{environment.get('name', 'error')}.updated_at"
            ] = environment.get("updated_at", False)
        return env_dict

    df = pd.concat(
        [
            df,
            df[environment_column]
            .apply(environments_to_dict)
            .apply(pd.Series)
            .fillna(False)
            .add_prefix(f"{environment_column}."),
        ],
        axis=1,
    )

    return df


def lookup_first_environment(
    df: pd.DataFrame, columns: Optional[List[str]] = None
) -> pd.Series:
    """Look up the first enabled environment in the DataFrame.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame with 'tenant.environments' or 'environments' column.

    Returns
    -------
    pd.Series
        Series with the first enabled environment name.
    """
    if df.empty:
        return df

    df = df.copy()

    if not columns:
        columns = ["tenant.environments", "environments"]

    environment_column = None
    for column in columns:
        if column in df.columns:
            environment_column = column
            break

    if not environment_column:
        raise ValueError(f"{columns} columns not found in DataFrame")

    def first_environment(environments: List[Dict[str, Any]]) -> Optional[str]:
        """Retun the first enabled environment name.

        Parameters
        ----------
        environments : List[Dict[str, Any]]
            Tenants API environments dictionary list.

        Returns
        -------
        Optional[str]
            Environment name, as found in the Tenants API.
        """
        if not environments:
            log.error("Environments data is empty or None.")
            return None

        try:
            for environment in environments:
                if environment["enabled"]:
                    return environment["name"]
        except KeyError as exc:
            log.error("KeyError in environment data: %s", exc)
            return None

        log.info("No enabled environment found.")
        return None

    return df[environment_column].apply(first_environment)


def lookup_tenants(
    df: pd.DataFrame, region: str, tenant_id_column: Optional[str] = None
) -> pd.DataFrame:
    """Look up tenants and correlate to the input dataframe.

    Parameters
    ----------
    df : pd.DataFrame
        Dataframe with tenant_id or tenant.id column
    region : str
        Taegis Region

    Returns
    -------
    pd.DataFrame
        Dataframe with correlated tenant.
        New columns will be prefixed with 'tenant.'
        Overlapping columns will be suffixed with '.lookup_tenants'

    Raises
    ------
    ValueError
        If the no valid tenant id column is found.
    """
    if df.empty:
        return df

    if not tenant_id_column:
        tenant_id_column = get_tenant_id_column(df)

    if not tenant_id_column in df.columns:
        raise ValueError(f"{tenant_id_column} not found in DataFrame columns.")

    if [
        column
        for column in df.columns
        if column.startswith("tenant.") and column != "tenant.id"
    ]:
        log.debug("Tenant columns already exist in DataFrame.")
        return df

    service = get_service(environment=region)

    max_results = 1000
    page_number = 1

    ids = df[tenant_id_column].unique().tolist()

    log.debug(f"Polling page: {page_number}")
    result = service.tenants.query.tenants(
        TenantsQuery(
            max_results=max_results,
            page_num=page_number,
            ids=ids,
        )
    )

    results = [result]

    while result.has_more:
        page_number += 1
        log.debug(f"Polling page: {page_number}")

        result = service.tenants.query.tenants(
            TenantsQuery(
                max_results=max_results,
                page_num=page_number,
                ids=ids,
            )
        )

        results.append(result)

    tenants_df = to_dataframe(
        [asdict(tenant) for result in results for tenant in result.results or []]
    ).add_prefix("tenant.")

    return pd.merge(
        df,
        tenants_df,
        left_on=tenant_id_column,
        right_on="tenant.id",
        how="left",
        suffixes=(None, ".lookup_tenants"),
    )



"""Pandas utility functions for correlating, transforming and pivoting on subjects in Taegis."""

import logging

from typing import Optional, List
import pandas as pd
from dataclasses import asdict

from taegis_magic.core.service import get_service
from taegis_magic.core.utils import to_dataframe
from taegis_magic.pandas.utils import chunk_list

log = logging.getLogger(__name__)


def lookup_users(
    df: pd.DataFrame,
    region: str,
    tenant_id: Optional[str] = None,
    user_id_columns: Optional[List[str]] = None,
    merge_ons: Optional[List[str]] = None,
) -> pd.DataFrame:
    """Look up users and correlate to the input dataframe.

    Parameters
    ----------
    df : pd.DataFrame
        Dataframe
    region : str
        Taegis Region
    tenant_id : Optional[str], optional
        Tenant ID to use for the lookup, by default None.
    user_id_columns : Optional[List[str]], optional
        List of user id columns to look up, by default None.
        If None, defaults to ['created_by', 'updated_by'].
    merge_ons : Optional[List[str]], optional
        List of user id fields to merge on, by default None.
        If None, defaults to ['user.id', 'user.idp_user_id', 'user.user_id']

    Returns
    -------
    pd.DataFrame
        Dataframe with correlated User.
        New columns will be prefixed with 'user.'
        Overlapping columns will be suffixed with '.lookup_users'

    Raises
    ------
    ValueError
        If the no valid user id column is found.
    """
    if df.empty:
        return df

    if not user_id_columns:
        user_id_columns = ["created_by", "updated_by"]

    if not merge_ons:
        merge_ons = ["user.id", "user.idp_user_id", "user.user_id"]

    for column in user_id_columns.copy():
        if not column in df.columns:
            log.error(f"Column {column} not found in dataframe")
            user_id_columns.remove(column)

    if not user_id_columns:
        log.error("No valid user id columns found in dataframe")
        return df

    if [
        column
        for column in df.columns
        for user_column in user_id_columns
        if column.startswith(f"{user_column}.user.")
        and column != f"{user_column}.user.id"
    ]:
        log.debug("User columns already exist in DataFrame.")
        return df

    service = get_service(tenant_id=tenant_id, environment=region)

    ids = set()
    for column in user_id_columns:
        ids.update(df[column].dropna().unique().tolist())
    ids = list(ids)

    results = []
    for chunk in chunk_list(ids, 100):
        log.debug(f"Correlating users: {chunk}")
        # need to access deprecated user_id field
        with service(
            output="""
            error
            user { 
                id idp_user_id user_id 
                created_at updated_at created_by 
                updated_by last_login last_ip invited_date 
                registered_date deactivated_date status 
                status_localized email email_normalized 
                family_name given_name phone_number 
                phone_extension secondary_phone_number 
                secondary_phone_extension environments 
                timezone tenant_status tenant_status_localized 
                entitlement_channel allowed_entitlement_channels 
                masked community_role is_scwx is_partner 
                pre_verified 
                accessible_tenants { 
                    id name name_normalized enabled allow_response_actions 
                    actions_approver expires_at is_partner is_organization partner 
                    parent 
                    environments { name enabled } 
                    labels { id tenant_id name value owner_partner_tenant_id } 
                    services { id name description } 
                } 
                role_assignments { 
                    id tenant_id role_id deactivated role_name role_display_name 
                    expires_at created_at updated_at created_by updated_by allowed_environments 
                } 
                eula { date version } 
                preferred_language 
            }
        """
        ):
            result = service.users.query.search_tdrusers_by_ids(user_ids=chunk)
        results.extend(result)

    users_df = to_dataframe([asdict(result) for result in results])

    for column in user_id_columns:
        dfs = []
        column_dataframe = df.copy()
        column_users = users_df.copy().add_prefix(f"{column}.")

        for merge_on in merge_ons:
            log.debug(f"Merge users for column: {column} with right_on: {merge_on}")

            merge_dataframe = pd.merge(
                column_dataframe,
                column_users,
                left_on=column,
                right_on=[f"{column}.{merge_on}"],
                how="left",
                suffixes=(None, ".lookup_users"),
            )

            # filter matched rows from merge
            filtered_dataframe = merge_dataframe[
                ~merge_dataframe[f"{column}.{merge_on}"].isna()
            ]
            if not filtered_dataframe.empty:
                dfs.append(filtered_dataframe)
                # filter out matched rows from df
                column_dataframe = column_dataframe[
                    ~column_dataframe[column].isin(
                        filtered_dataframe[f"{column}.{merge_on}"]
                    )
                ]

        if not column_dataframe.empty:
            log.debug(f"Adding column dataframe: {column_dataframe['id']}")
            dfs.append(column_dataframe)

        df = pd.concat(dfs, ignore_index=True) if dfs else pd.DataFrame()

    return df



"""Pandas functions for Alerts DataFrames."""

import logging
import re
import time
from typing import Dict, List, Optional

import pandas as pd
from numpy import add
from taegis_magic.commands.alerts import (
    AlertsResultsNormalizer,
    alerts_service_search_with_events,
)
from taegis_magic.core.normalizer import TaegisResultsNormalizer
from taegis_magic.core.service import get_service
from taegis_magic.pandas.utils import chunk_list, coalesce_columns

from taegis_sdk_python.services.alerts.types import (
    ResolutionStatus,
    SearchRequestInput,
    TimestampInput,
    UpdateResolutionRequestInput,
)

log = logging.getLogger(__name__)


def convert_alert_timestamps(
    df: pd.DataFrame, format_: str = "%Y-%m-%dT%H:%M:%SZ"
) -> pd.DataFrame:
    """Takes an Alerts dataframe and converts all the metadata time columns
    into a readable time format. All columns with .nanos will be dropped to reduce
    the amount of columns in the dataframe.

    Parameters
    ----------
    df : pd.DataFrame
        Alerts Dataframe
    format : str, optional
        Datetime string format to be used, by default "%Y-%m-%dT%H:%M:%SZ"

    Returns
    -------
    pd.DataFrame
        Returns Alerts Dataframe with added taegis_magic timestamp columns
    """

    if df.empty:
        return df

    df = df.copy()

    for column in [
        column
        for column in df.columns
        if column.endswith(".seconds") and not column.startswith("taegis_magic.")
    ]:
        try:
            df[f"taegis_magic.{column}"] = pd.to_datetime(
                df[column], errors="ignore", unit="s"
            ).dt.strftime(format_)
        except Exception as exc:
            log.error(exc)
            continue

        df[f"taegis_magic.{column}"] = df[f"taegis_magic.{column}"].fillna("N/A")

    return df


def inflate_raw_events(df: pd.DataFrame) -> pd.DataFrame:
    """Takes a DataFrame containing Alert (V2) results
    and concats the nested JSON blob containing the
    source event data as columns. Returns a DataFrame.

    Parameters
    ----------
    df : pd.DataFrame
        Alerts (V2) DataFrame

    Returns
    -------
    pd.DataFrame
        Returns an Alerts (V2) Dataframe with exploded events data

    Raises
    ------
    ValueError
        Dataframe does not contain an event_ids column
    """

    if df.empty:
        return df

    if "event_ids" not in df.columns:
        raise ValueError("Dataframe does not contain an event_ids column")

    if not any(df.columns.str.startswith("event_data.")):
        df = df.explode("event_ids").reset_index(drop=True)

        event_df = pd.json_normalize(df["event_ids"]).dropna(axis=1, how="all")

        no_prefix = [col for col in event_df.columns if "event_data." not in col]

        if no_prefix:
            for column in no_prefix:
                event_df = event_df.rename({column: f"event_data.{column}"}, axis=1)

        return pd.concat(
            [
                df,
                event_df,
            ],
            axis=1,
        )

    return df


def get_alerts_from_aggregation(
    df: pd.DataFrame,
    region: Optional[str] = None,
    tenant: Optional[str] = None,
    earliest: str = "-7d",
    latest: Optional[str] = None,
    limit: int = 100,
    additional_operators: Optional[List[str]] = None,
) -> pd.DataFrame:
    """This function takes a DataFrame of aggregate alerts and retrieves
    corresponding non-aggregate data for the same alerts. The main use-case
    is reviewing the aggregate alert data in an interactive table widget
    (such as IPyDataGrid), then use this function to pull down the
    non-aggregate alerts to look more closely at their contents.

    Parameters
    ----------
    df : pd.DataFrame
        Pandas DataFrame of Alerts Aggregation
    region : str, optional
        Taegis environment, by default "charlie", options: "charlie", "delta", "echo"
    tenant : str, optional
        Which tenant_id to use for the alerts query
    earliest : str, optional
        Earliest timeframe in the alert query, defaults to -7d.
    latest : str, optional
        Latest timeframe in the alert query, defaults to None.
    limit : int, optional
        Alert results limit, defaults to 100.
    additional_operators : Optional[List[str]], optional
        Additional operators to add to the query, by default None.

    Returns
    -------
    pd.DataFrame
        Alerts Dataframe

    Raises
    ------
    ValueError
        If the Dataframe contained no data to populate sub-queries for the aggregate alerts query
    """

    cols = [
        column
        for column in df.columns
        if column not in ["count"] and re.match("^[^0-9]", column) is not None
    ]

    sub_query = []
    single_quote = "'"
    replacement = "\\'"
    for _, row in df.iterrows():
        row_query = [
            (
                f"{col} = '{row[col]}'"
                if not str(row[col]).find("'") > -1
                else f"{col} = e'{str(row[col]).replace(single_quote, replacement)}'"
            )
            for col in cols
            if col in row
        ]

        sub_query.append("(" + " AND ".join(row_query) + ")")

    if not sub_query:
        raise ValueError(
            "No sub-queries in the alerts query WHERE statement. Please look to see if your dataframe has aggregate alert data."
        )

    unique_sub_queries = []

    for x in sub_query:
        if x not in unique_sub_queries:
            unique_sub_queries.append(x)

    sub_query_string = " OR \n".join(x for x in unique_sub_queries)

    additional_operators_string = ""
    if additional_operators:
        additional_operators_string = " AND \n ".join(
            [f"({operator})" for operator in additional_operators]
        )
        additional_operators_string = f"AND \n{additional_operators_string}"

        for op in additional_operators:
            for col in cols:
                if col in op:
                    log.warning(
                        f"Column {col} is already in the sub-query. Please check your additional operators."
                    )

    query = f"""
    FROM alert
    WHERE ({sub_query_string})
    {additional_operators_string}
    {f"EARLIEST={earliest}" if earliest else ""}
    {f"LATEST={latest}" if latest else ""}
    """

    service = get_service(environment=region, tenant_id=tenant)

    try:
        response = alerts_service_search_with_events(
            service,
            SearchRequestInput(
                cql_query=query,
                offset=0,
                limit=limit,
            ),
        )
    except Exception as exc:
        log.error(exc)
        return pd.DataFrame()

    normalized_results = AlertsResultsNormalizer(
        raw_results=[response],
        service="alerts",
        tenant_id=tenant,
        region=region,
        arguments={
            "query": query,
            "tenant": tenant,
            "region": region,
            "earliest": earliest,
            "latest": latest,
        },
    )

    return pd.json_normalize(
        normalized_results.results,
        max_level=3,
    )


def inflate_third_party_details(df: pd.DataFrame) -> pd.DataFrame:
    """Expands `third_party_details` column of a DataFrame and returns
    a new DataFrame where the content of `third_party_details` is
    represented as columns.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame containing an `third_party_details` column
    Returns
    -------
    pd.DataFrame
        DataFrame with `third_party_details` columns appended

    Example
    -------
    Example::

        %%xdr ql alerts get --assign third_party_alerts --cache
        FROM alert
        WHERE third_party_details IS NOT NULL
        EARLIEST=-1d | head 25

        third_party_alerts_inflated = third_party_alerts.pipe(inflate_third_party_details)

    """

    def third_party_details_to_dict(third_party_details: List[Dict[str, str]]) -> Dict:
        parsed_third_party_details = []
        for detail in third_party_details:
            kv_pairs = detail.get("generic", {}).get("generic", {}).get("record", [])
            for kv_pair in kv_pairs:
                parsed_third_party_details.append(tuple(kv_pair.values()))
        return dict(parsed_third_party_details)

    if "third_party_details" in df.columns:
        if df["third_party_details"].any():
            return pd.concat(
                [
                    df,
                    df["third_party_details"]
                    .apply(third_party_details_to_dict)
                    .apply(pd.Series)
                    .add_prefix("third_party_details."),
                ],
                axis=1,
            )

        log.warning("third_party_details column contains no data to be inflated.")
        return df

    log.warning("Dataframe did not contain a third_party_details column.")
    return df


def severity_rounded_and_category(
    df: pd.DataFrame, severity_columns: Optional[List[str]] = None
) -> pd.DataFrame:
    """Converts the Taegis alert severity and creates two new taegis_magic. columns.
    One column will convert the severity to a numeric value, and round it to two decimal places.
    The second column will then apply a severity category such as Informational,
    Critical, High, Medium, and Low.

    Parameters
    ----------
    df : pd.DataFrame
        Taegis Alerts Dataframe that contains a severity column.

    Returns
    -------
    pd.DataFrame
        Returns a dataframe containing two taegis_magic. columns if the passed in dataframe contains a
        severity column. If not, the original dataframe is returned.
    """
    if df.empty:
        return df

    df = df.copy()

    def bucket_severity_scores(severity):
        if severity >= 0.8:
            return "Critical"
        elif severity < 0.8 and severity >= 0.6:
            return "High"
        elif severity < 0.6 and severity >= 0.4:
            return "Medium"
        elif severity < 0.4 and severity >= 0.2:
            return "Low"
        elif severity < 0.2:
            return "Informational"
        else:
            raise ValueError

    if severity_columns is None:
        # sorted by priority of field
        severity_columns = ["metadata.severity", "severity"]

    log.debug("Checking severity identifier columns...")
    valid_severity_columns = []

    for severity_col in severity_columns:
        if severity_col in df.columns:
            valid_severity_columns.append(severity_col)
            log.debug(f"Found severity identifier column: {severity_col}...")

    if not valid_severity_columns:
        raise ValueError(
            f"DataFrame does not contain a vaild severity column: {valid_severity_columns}"
        )
    df["taegis_magic.severity"] = (
        coalesce_columns(df, valid_severity_columns).apply(pd.to_numeric).round(2)
    )
    df["taegis_magic.severity_category"] = df["taegis_magic.severity"].apply(
        bucket_severity_scores
    )
    return df


def provide_feedback(
    df: pd.DataFrame,
    environment: str,
    status: ResolutionStatus,
    reason: str,
) -> pd.DataFrame:
    """
    Resolve a DataFrame of Alerts.

    Parameters
    ----------
    df : pd.DataFrame
        Alerts DataFrame
    status : ResolutionStatus
        Status to resolve alerts
    reason : str, optional
        Reason for resolution status

    Returns
    -------
    pd.DataFrame
        Alerts DataFrame
    """

    service = get_service(environment=environment)

    if "tenant_id" in df.columns:
        tenant_identifier = "tenant_id"
    elif "tenant.id" in df.columns:
        tenant_identifier = "tenant.id"
    else:
        raise ValueError("DataFrame does not contain a valid tenant identifier")

    tenants = set(df[tenant_identifier].to_list())

    for tenant in tenants:
        alert_ids = df[df[tenant_identifier] == tenant]["id"].unique()

        with service(tenant_id=tenant):
            for chunk in chunk_list(alert_ids, 250):
                service.alerts.mutation.alerts_service_update_resolution_info(
                    UpdateResolutionRequestInput(
                        alert_ids=chunk,
                        resolution_status=status,
                        reason=reason,
                        requested_at=TimestampInput(seconds=int(time.time())),
                    )
                )

    return df


def normalize_creator_name(
    df: pd.DataFrame, columns: Optional[List[str]] = None, region: Optional[str] = None
) -> pd.DataFrame:
    """
    Normalize creator ids to pretty names.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame with `creator` or `metadata.creator.detector.detector_id` fields
    columns : Optional[List[str]], optional
        User provided columns, by default None
    region : Optional[str], optional
        Region to query, by default None

    Returns
    -------
    pd.DataFrame
        DataFrame with pretty names.

    Raises
    ------
    KeyError
        No valid columns to correlate creator names.
    """
    df = df.copy()

    if df.empty:
        return df

    if not columns:
        columns = ["creator", "metadata.creator.detector.detector_id"]

    column = None
    for c in columns:
        if c in df.columns:
            column = c
            break

    if not column:
        raise KeyError("No valid columns to correlate creator names.")

    service = get_service(environment=region)

    detector_results = service.detector_registry.query.detectors()

    df["taegis_magic.creator.display_name"] = df.apply(
        lambda x: next(
            iter(
                [
                    detector.display_name
                    for detector in detector_results
                    if detector.creator_name == x[column]
                ]
                or [x[column]]
            )
        ),
        axis=1,
    )

    return df



"""Taegis Magic users commands."""

import inspect
import logging
from dataclasses import asdict, field
from enum import Enum
from pprint import pprint
from typing import Any, List, Optional

import typer
from taegis_magic.commands.clients import ROLE_MAP, Roles
from taegis_magic.core.log import tracing
from taegis_magic.core.normalizer import TaegisResultsNormalizer

from taegis_magic.core.service import get_service
from taegis_sdk_python.services.users.types import (
    InviteUsersResponse,
    SupportPinDetails,
    TDRUser,
    TDRUserInviteInput,
    TDRUsersSearchInput,
    TDRUsersSearchResults,
    TDRUserSupportPin,
    TDRUserTrialInviteInput,
    TDRUsersLanguage,
)
from typing_extensions import Annotated

app = typer.Typer(help="Taegis User Commands.")

log = logging.getLogger(__name__)


class TaegisUserResultsNormalizer(TaegisResultsNormalizer):
    raw_results: Any = field(default=None)

    @property
    def results(self):
        return [asdict(self.raw_results)]


class TaegisUsersResultsNormalizer(TaegisResultsNormalizer):
    raw_results: List[Any] = field(default_factory=list)

    @property
    def results(self):
        return [asdict(r) for r in self.raw_results]


class TaegisUserSearchResultsNormalizer(TaegisResultsNormalizer):
    raw_results: TDRUsersSearchResults = field(
        default_factory=lambda: TDRUsersSearchResults()
    )

    @property
    def results(self):
        return [asdict(user) for user in self.raw_results.results]


@app.command()
@tracing
def search(
    email: Annotated[Optional[List[str]], typer.Option()] = None,
    role: Annotated[Optional[List[Roles]], typer.Option()] = None,
    status: Optional[str] = None,
    tenant_status: Optional[str] = None,
    name: Optional[str] = None,
    tenant: Optional[str] = None,
    region: Optional[str] = None,
):
    """Search for users."""
    service = get_service(environment=region, tenant_id=tenant)

    if email:
        email_partials = [e for e in email if "%" in e]

        try:
            email_match = next(iter(email_partials))
        except StopIteration:
            email_match = None

        emails = [e for e in email if e not in email_partials]

        if email_match:
            if len(email_partials) > 1:
                log.warning(
                    "Only a single '%' email filter may be used at time.  Dropping remaining partials..."
                )

            if emails:
                log.warning(
                    "Cannot search for partial emails and full emails in the same search.  Dropping full emails..."
                )

            emails = None
    else:
        email_match = None
        emails = None

    if role:
        roles = [ROLE_MAP[r.value] for r in role]
    else:
        roles = None

    results = service.users.query.tdr_users_search(
        TDRUsersSearchInput(
            email=email_match,
            emails=emails,
            role_ids=roles,
            status=status,
            tenant_status=tenant_status,
            name=name,
        )
    )

    normalized_results = TaegisUserSearchResultsNormalizer(
        raw_results=results,
        service="users",
        tenant_id=service.tenant_id,
        region=service.environment,
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


@app.command()
@tracing
def support_pin(region: Optional[str] = None):
    """Get a PIN for Taegis support calls."""
    service = get_service(environment=region)

    results = service.users.query.get_support_pin()

    normalized_results = TaegisUserResultsNormalizer(
        raw_results=results,
        service="users",
        tenant_id=service.tenant_id,
        region=service.environment,
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


@app.command()
@tracing
def verify_support_pin(pin: str, region: Optional[str] = None):
    """Verify users associated with PIN for Taegis support calls."""
    service = get_service(environment=region)

    results = service.users.query.get_support_pin_verification(pin)

    normalized_results = TaegisUsersResultsNormalizer(
        raw_results=results,
        service="users",
        tenant_id=service.tenant_id,
        region=service.environment,
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


@app.command()
@tracing
def current_user(region: Optional[str] = None):
    """Get current user information."""
    service = get_service(environment=region)

    results = service.users.query.current_tdruser()

    normalized_results = TaegisUserResultsNormalizer(
        raw_results=results,
        service="users",
        tenant_id=service.tenant_id,
        region=service.environment,
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


@app.command()
@tracing
def invite_users(
    email: Annotated[List[str], typer.Option()],
    role: Annotated[Roles, typer.Option()] = Roles.analyst,
    language: Annotated[TDRUsersLanguage, typer.Option()] = TDRUsersLanguage.EN,
    tenant: Annotated[Optional[str], typer.Option()] = None,
    region: Annotated[Optional[str], typer.Option()] = None,
):
    """Invite users to Taegis."""
    service = get_service(environment=region, tenant_id=tenant)

    results = service.users.mutation.invite_tdr_users(
        invites=[
            TDRUserInviteInput(email=e, role_id=ROLE_MAP[role.value], language=language)
            for e in email
        ]
    )

    normalized_results = TaegisUsersResultsNormalizer(
        raw_results=results,
        service="users",
        tenant_id=service.tenant_id,
        region=service.environment,
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


@app.command()
@tracing
def invite_trial_user(
    email: Annotated[str, typer.Option()],
    trial_tenant: Annotated[str, typer.Option()],
    language: Annotated[TDRUsersLanguage, typer.Option()] = TDRUsersLanguage.EN,
    tenant: Annotated[Optional[str], typer.Option()] = None,
    region: Annotated[Optional[str], typer.Option()] = None,
):
    """Invite users to Taegis."""
    service = get_service(environment=region, tenant_id=tenant)

    results = service.users.mutation.invite_trial_tdr_user(
        invites=TDRUserTrialInviteInput(
            email=email,
            tenant=trial_tenant,
            language=language,
        )
    )

    normalized_results = TaegisUsersResultsNormalizer(
        raw_results=results,
        service="users",
        tenant_id=service.tenant_id,
        region=service.environment,
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


@app.command()
@tracing
def validate_support_pin(email: str, pin: str, region: Optional[str] = None):
    """Validate a user's support pin."""
    service = get_service(environment=region)

    results = service.users.mutation.validate_support_pin(
        email=email,
        support_pin=pin,
    )

    normalized_results = TaegisUserResultsNormalizer(
        raw_results=results,
        service="users",
        tenant_id=service.tenant_id,
        region=service.environment,
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


if __name__ == "__main__":
    rv = app(standalone_mode=False)
    pprint(rv)



"""Taegis Magic tenants commands."""

import inspect
import logging
from dataclasses import asdict, dataclass, field
from typing import List, Optional

import typer
from dataclasses_json import dataclass_json
from taegis_magic.core.log import tracing
from taegis_magic.core.normalizer import TaegisResultsNormalizer
from taegis_magic.core.service import get_service
from taegis_magic.core.utils import remove_output_node
from taegis_sdk_python import (
    GraphQLNoRowsInResultSetError,
    build_output_string,
)
from taegis_sdk_python.services.rules.types import (
    Rule,
    RuleEventType,
    RuleQueryKind,
    RuleType,
    SearchRulesInput,
    SearchRulesOutput,
    RuleInput,
    RuleQLFilterInput,
    RuleFilterInput,
)
from typing_extensions import Annotated
from enum import Enum
from click.exceptions import BadOptionUsage


log = logging.getLogger(__name__)

app = typer.Typer(help="Taegis Rules Commands.")
create_rule = typer.Typer(help="Create a Custom Alerting or Suppression Rules.")
app.add_typer(create_rule, name="create")


class RuleSeverity(str, Enum):
    """Rule Severity Options."""

    low = "low"
    medium = "medium"
    high = "high"
    critical = "critical"


RULE_SEVERTIY_MAP = {
    "low": 0.2,
    "medium": 0.4,
    "high": 0.6,
    "critical": 0.8,
}


class RuleSuppressionKey(str, Enum):
    """Suppression Rule keys."""

    alert_title = "alert_title"
    detector = "detector"
    entity = "entity"


RULE_SUPPRESSION_KEY_MAP = {
    "alert_title": "scwx.observation_v2.title",
    "detector": "scwx.observation_v2.detector_id",
    "entity": "scwx.observation_v2.entity",
}


def multiple_rules_output():
    return remove_output_node(
        build_output_string(Rule),
        "generativeAIRuleExplain",
    )


@dataclass_json
@dataclass
class TaegisRuleNormalizer(TaegisResultsNormalizer):
    """Taegis Single Rule normalizer."""

    raw_results: Rule = field(default_factory=lambda: Rule())

    @property
    def results(self):
        return [asdict(self.raw_results)]


@dataclass
class TaegisRulesNormalizer(TaegisResultsNormalizer):
    """Taegis Multiple Rule normalizer."""

    raw_results: List[Rule] = field(default_factory=list)

    @property
    def results(self):
        return [asdict(rule) for rule in self.raw_results]


@dataclass
class TaegisRulesSearchNormalizer(TaegisResultsNormalizer):
    """Taegis Search Rule normalizer."""

    raw_results: List[SearchRulesOutput] = field(
        default_factory=lambda: [SearchRulesOutput()]
    )

    @property
    def results(self):
        return [
            asdict(rule)
            for result in self.raw_results or []
            for rule in result.rules or []
        ]


@app.command(name="type")
@tracing
def rules_type(
    rule_type: Annotated[RuleType, typer.Option()] = RuleType.QL,
    tenant: Optional[str] = None,
    region: Optional[str] = None,
):
    """Taegis rules by type."""
    service = get_service(environment=region, tenant_id=tenant)

    page = 1
    per_page = 500
    rules = []

    log.info(f"Getting page: {page}")
    try:
        with service(output=multiple_rules_output()):
            results = service.rules.query.rules(
                page=page, count=per_page, rule_type=rule_type
            )
    except GraphQLNoRowsInResultSetError:  # pragma: no cover
        results = []
    log.debug(f"Results returned: {len(results)}")
    rules.extend(results)

    while len(results) == per_page:
        page += 1
        log.info(f"Getting page: {page}")
        try:
            with service(output=multiple_rules_output()):
                results = service.rules.query.rules(
                    page=page, count=per_page, rule_type=rule_type
                )
        except GraphQLNoRowsInResultSetError:  # pragma: no cover
            results = []
        log.debug(f"Results returned: {len(results)}")
        rules.extend(results)

    normalized_results = TaegisRulesNormalizer(
        raw_results=rules,
        service="rules",
        tenant_id=service.tenant_id,
        region=service.environment,
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


@app.command(name="all")
@tracing
def rules_all(
    tenant: Optional[str] = None,
    region: Optional[str] = None,
):
    """Taegis all rules."""
    service = get_service(environment=region, tenant_id=tenant)

    page = 1
    per_page = 500
    rules = []

    log.info(f"Getting page: {page}")
    try:
        with service(output=multiple_rules_output()):
            results = service.rules.query.all_rules(page=page, count=per_page)
    except GraphQLNoRowsInResultSetError:  # pragma: no cover
        results = []
    log.debug(f"Results returned: {len(results)}")
    rules.extend(results)

    while len(results) == per_page:
        page += 1
        log.info(f"Getting page: {page}")
        try:
            with service(output=multiple_rules_output()):
                results = service.rules.query.all_rules(page=page, count=per_page)
        except GraphQLNoRowsInResultSetError:  # pragma: no cover
            results = []
        log.debug(f"Results returned: {len(results)}")
        rules.extend(results)

    normalized_results = TaegisRulesNormalizer(
        raw_results=rules,
        service="rules",
        tenant_id=service.tenant_id,
        region=service.environment,
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


@app.command(name="suppression")
@tracing
def rules_suppression(
    kind: RuleQueryKind = RuleQueryKind.ALL,
    tenant: Optional[str] = None,
    region: Optional[str] = None,
):
    """Taegis suppression rules."""
    service = get_service(environment=region, tenant_id=tenant)

    page = 1
    per_page = 500
    rules = []

    log.info(f"Getting page: {page}")
    try:
        results = service.rules.query.suppression_rules(
            page=page,
            count=per_page,
            kind=kind,
        )
    except GraphQLNoRowsInResultSetError:  # pragma: no cover
        results = []
    log.debug(f"Results returned: {len(results)}")
    rules.extend(results)

    while len(results) == per_page:
        page += 1
        log.info(f"Getting page: {page}")
        try:
            results = service.rules.query.suppression_rules(
                page=page,
                count=per_page,
                kind=kind,
            )
        except GraphQLNoRowsInResultSetError:  # pragma: no cover
            results = []
        log.debug(f"Results returned: {len(results)}")
        rules.extend(results)

    normalized_results = TaegisRulesNormalizer(
        raw_results=rules,
        service="rules",
        tenant_id=service.tenant_id,
        region=service.environment,
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


@app.command(name="deleted")
@tracing
def rules_deleted(
    rule_type: Annotated[RuleType, typer.Option()] = RuleType.QL,
    tenant: Optional[str] = None,
    region: Optional[str] = None,
):
    """Taegis deleted rules."""
    service = get_service(environment=region, tenant_id=tenant)

    page = 1
    per_page = 500
    rules = []

    log.info(f"Getting page: {page}")
    try:
        with service(output=multiple_rules_output()):
            results = service.rules.query.deleted_rules(
                page=page,
                count=per_page,
                rule_type=rule_type,
            )
    except GraphQLNoRowsInResultSetError:  # pragma: no cover
        results = []
    log.debug(f"Results returned: {len(results)}")
    rules.extend(results)

    while len(results) == per_page:
        page += 1
        log.info(f"Getting page: {page}")
        try:
            with service(output=multiple_rules_output()):
                results = service.rules.query.deleted_rules(
                    page=page,
                    count=per_page,
                    rule_type=rule_type,
                )
        except GraphQLNoRowsInResultSetError:  # pragma: no cover
            results = []
        log.debug(f"Results returned: {len(results)}")
        rules.extend(results)

    normalized_results = TaegisRulesNormalizer(
        raw_results=rules,
        service="rules",
        tenant_id=service.tenant_id,
        region=service.environment,
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


@app.command(name="filters")
@tracing
def rules_event_type(
    event_type: Annotated[RuleEventType, typer.Option()],
    rule_type: Annotated[RuleType, typer.Option()] = RuleType.REGEX,
    tenant: Optional[str] = None,
    region: Optional[str] = None,
):
    """Taegis all rules."""
    service = get_service(environment=region, tenant_id=tenant)

    page = 1
    per_page = 500
    rules = []

    log.info(f"Getting page: {page}")
    try:
        with service(output=multiple_rules_output()):
            results = service.rules.query.rules_for_event(
                page=page, count=per_page, event_type=event_type, rule_type=rule_type
            )
    except GraphQLNoRowsInResultSetError:  # pragma: no cover
        results = []
    log.debug(f"Results returned: {len(results)}")
    rules.extend(results)

    while len(results) == per_page:
        page += 1
        log.info(f"Getting page: {page}")
        try:
            with service(output=multiple_rules_output()):
                results = service.rules.query.rules_for_event(
                    page=page,
                    count=per_page,
                    event_type=event_type,
                    rule_type=rule_type,
                )
        except GraphQLNoRowsInResultSetError:  # pragma: no cover
            results = []
        log.debug(f"Results returned: {len(results)}")
        rules.extend(results)

    normalized_results = TaegisRulesNormalizer(
        raw_results=rules,
        service="rules",
        tenant_id=service.tenant_id,
        region=service.environment,
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


@app.command(name="rule")
@tracing
def rules_rule(
    rule_id: Annotated[str, typer.Option()],
    tenant: Optional[str] = None,
    region: Optional[str] = None,
):
    """Taegis rule by id."""
    service = get_service(environment=region, tenant_id=tenant)

    try:
        results = service.rules.query.rule(id_=rule_id)
    except GraphQLNoRowsInResultSetError:  # pragma: no cover
        results = Rule()

    normalized_results = TaegisRuleNormalizer(
        raw_results=results,
        service="rules",
        tenant_id=service.tenant_id,
        region=service.environment,
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


@app.command("changes-since")
@tracing
def rules_changes_since(
    timestamp: Annotated[str, typer.Option(help="YYYY-DD-MMTHH:MM:SSZ")],
    event_type: Annotated[Optional[RuleEventType], typer.Option()] = None,
    rule_type: Annotated[RuleType, typer.Option()] = RuleType.REGEX,
    tenant: Optional[str] = None,
    region: Optional[str] = None,
):
    """Taegis rule changed by timestamp."""
    service = get_service(environment=region, tenant_id=tenant)

    try:
        results = service.rules.query.changes_since(
            timestamp=timestamp,
            event_type=event_type,
            rule_type=rule_type,
        )
    except GraphQLNoRowsInResultSetError:  # pragma: no cover
        results = []

    normalized_results = TaegisRulesNormalizer(
        raw_results=results,
        service="rules",
        tenant_id=service.tenant_id,
        region=service.environment,
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


@app.command("search")
@tracing
def rules_search(
    cell: Annotated[str, typer.Option()],
    tenant: Optional[str] = None,
    region: Optional[str] = None,
):
    """Taegis rule changed by timestamp."""
    service = get_service(environment=region, tenant_id=tenant)

    page = 1
    per_page = 500
    rules = []

    log.info(f"Getting page: {page}")
    try:
        with service(
            output=remove_output_node(
                build_output_string(SearchRulesOutput), "generativeAIRuleExplain"
            )
        ):
            results = service.rules.query.search_rules(
                SearchRulesInput(query=cell),
                page=page,
                count=per_page,
            )
    except GraphQLNoRowsInResultSetError:  # pragma: no cover
        results = []
    log.debug(f"Results returned: {len(results.rules)}")
    rules.append(results)

    while len(results.rules) == per_page:
        page += 1
        log.info(f"Getting page: {page}")
        try:
            with service(
                output=remove_output_node(
                    build_output_string(SearchRulesOutput), "generativeAIRuleExplain"
                )
            ):
                results = service.rules.query.search_rules(
                    SearchRulesInput(query=cell),
                    page=page,
                    count=per_page,
                )
        except GraphQLNoRowsInResultSetError:  # pragma: no cover
            results = []
        log.debug(f"Results returned: {len(results.rules)}")
        rules.append(results)

    normalized_results = TaegisRulesSearchNormalizer(
        raw_results=rules,
        service="rules",
        tenant_id=service.tenant_id,
        region=service.environment,
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


@app.command(name="enable")
@tracing
def rules_enable(
    rule_id: Annotated[str, typer.Option()],
    tenant: Optional[str] = None,
    region: Optional[str] = None,
):
    """Enable a Taegis rule."""

    service = get_service(environment=region, tenant_id=tenant)

    results = service.rules.mutation.enable_rule(id_=rule_id)

    normalized_results = TaegisRuleNormalizer(
        raw_results=results,
        service="rules",
        tenant_id=service.tenant_id,
        region=service.environment,
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


@app.command(name="disable")
@tracing
def rules_disable(
    rule_id: Annotated[str, typer.Option()],
    tenant: Optional[str] = None,
    region: Optional[str] = None,
):
    """Disable a Taegis rule."""

    service = get_service(environment=region, tenant_id=tenant)

    results = service.rules.mutation.disable_rule(id_=rule_id)

    normalized_results = TaegisRuleNormalizer(
        raw_results=results,
        service="rules",
        tenant_id=service.tenant_id,
        region=service.environment,
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


@create_rule.command(name="custom")
@tracing
def rules_create_custom_rule(
    cell: Annotated[str, typer.Option()],
    name: Annotated[str, typer.Option()],
    description: Annotated[str, typer.Option()],
    severity: Annotated[RuleSeverity, typer.Option()],
    mitre_category: Annotated[Optional[List[str]], typer.Option()] = None,
    tenant: Optional[str] = None,
    region: Optional[str] = None,
):
    """Taegis Create Custom QL Rule."""
    service = get_service(environment=region, tenant_id=tenant)

    results = service.rules.mutation.create_custom_ql_rule(
        input_=RuleInput(
            name=name,
            description=description,
            severity=RULE_SEVERTIY_MAP[severity.value],
            create_alert=True,
            attack_categories=mitre_category,
        ),
        ql_filter=RuleQLFilterInput(query=cell),
    )

    normalized_results = TaegisRuleNormalizer(
        raw_results=results,
        service="rules",
        tenant_id=service.tenant_id,
        region=service.environment,
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


@create_rule.command(name="suppression")
@tracing
def rules_create_suppression_rule(
    name: Annotated[str, typer.Option()],
    description: Annotated[str, typer.Option()],
    key: Annotated[List[RuleSuppressionKey], typer.Option()],
    pattern: Annotated[
        List[str],
        typer.Option(help="Regex pattern match, ensure you escape literals (ie '\\.')"),
    ],
    inverted: Annotated[
        Optional[List[str]],
        typer.Option(
            help="required for all if a single filter is inverted, set to True/true, all other values will be False"
        ),
    ] = None,
    tenant: Optional[str] = None,
    region: Optional[str] = None,
):
    """Taegis Create Suppression Rule."""
    service = get_service(environment=region, tenant_id=tenant)

    if len(key) != len(pattern):  # pragma: no cover
        raise BadOptionUsage(..., "key and pattern lengths must match")

    if inverted and len(key) != len(inverted):  # pragma: no cover
        raise BadOptionUsage(..., "key and inverted lengths must match")

    if inverted:
        inverted = [True if i.lower() == "true" else False for i in inverted]
    else:
        inverted = [False for _ in key]

    entity_prefixes = service.rules.query.entity_prefixes()
    filters = []

    for k, p, i in zip(key, pattern, inverted):
        if k == RuleSuppressionKey.entity:
            if ":" not in p:  # pragma: no cover
                raise BadOptionUsage(
                    ...,
                    f"'entity' patterns must be formatted 'prefix:pattern', prefix must be in value: {list(entity_prefixes.keys())}",
                )

            entity_prefix = p.split(":", maxsplit=1)[0]

            if entity_prefix not in entity_prefixes:  # pragma: no cover
                raise BadOptionUsage(
                    ...,
                    f"'entity' patterns must be formatted 'prefix:pattern', prefix must be in value: {list(entity_prefixes.keys())}",
                )

        filters.append(
            RuleFilterInput(
                key=RULE_SUPPRESSION_KEY_MAP[k.value], pattern=p, inverted=i
            )
        )

    results = service.rules.mutation.create_custom_suppression_rule(
        input_=RuleInput(
            name=name,
            description=description,
        ),
        filters=filters,
    )

    normalized_results = TaegisRuleNormalizer(
        raw_results=results,
        service="rules",
        tenant_id=service.tenant_id,
        region=service.environment,
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results



"""Pandas functions for Asset Lookups in Event and Alert DataFrames."""

import logging
import pandas as pd
from typing import List, Optional
from dataclasses import asdict

from taegis_sdk_python import GraphQLService
from taegis_sdk_python.services.assets2.types import (
    AssetV2,
    AssetFilter,
    AssetWhereInputV2,
)
from taegis_magic.pandas.utils import chunk_list, get_tenant_id
from taegis_magic.core.service import get_service
from taegis_magic.core.utils import to_dataframe

log = logging.getLogger(__name__)


def get_assets(
    service: GraphQLService,
    filter: AssetFilter,
) -> List[AssetV2]:
    """Takes an asset filter to obtain a list of assets

    Parameters
    ----------
    service : GraphQLService
        Taegis SDK GraphQL service.
    filter : AssetFilter
        Asset filter that specifies what assets to return back.

    Returns
    -------
    List[AssetV2]
        Returns a list of AssetV2 objects.
    """
    all_assets = []

    response = service.assets2.query.assets_v2(
        filter_=filter,
    )

    all_assets.extend(response.assets)
    more_pages = response.page_info.has_next_page
    end_cursor = response.page_info.end_cursor

    while more_pages:
        response = service.assets2.query.assets_v2(
            after=end_cursor,
            filter_=filter,
        )
        more_pages = response.page_info.has_next_page
        end_cursor = response.page_info.end_cursor
        all_assets.extend(response.assets)

    return all_assets


def assets_from_list(
    service: GraphQLService,
    asset_list: List[str],
):
    """Takes a list of assets and returns a list of AssetV2 objects.

    Parameters
    ----------
    service : GraphQLService
        Taegis SDK GraphQL service object.
    asset_list : List[str]
        A list of Taegis host_ids to be used in the asset lookup.

    Returns
    -------
    List[AssetV2]
        Returns a list of AssetV2 objects.
    """
    asset_data = get_assets(
        service=service,
        filter=AssetFilter(
            where=AssetWhereInputV2(
                or_=[AssetWhereInputV2(host_id=x) for x in asset_list]
            )
        ),
    )

    return asset_data


def lookup_assets(
    df: pd.DataFrame, env: Optional[str] = None, region: Optional[str] = None
) -> pd.DataFrame:
    """Takes a Taegis pandas dataframe that contains host_ids and tenant_ids columns
    and preforms a assetv2 lookup using the Taegis SDK on the unique host_ids.

    Parameters
    ----------
    df : pd.DataFrame
        A pandas DataFrame that contains Taegis host_ids and tenant_ids.
    env : Optional[str], deprecated
        Taegis SDK Region/Environment that the asset lookup is for.  Defaults to US1.
    region : Optional[str]
        Taegis SDK Region/Environment that the asset lookup is for.  Defaults to US1.

    Returns
    -------
    pd.DataFrame
        Returns a pandas DataFrame with additional asset information columns.

    Raises
    ------
    ValueError
        If there are no valid host_id columns in the dataframe a value error will be raised.
    """
    if env:
        log.warning("The `env` parameter is deprecated. Please use `region` instead.")
        if not region:
            region = env

    df = df.copy()

    if df.empty:
        return df

    host_id_col = next(
        x for x in df.columns if "host_id" in x and not "asset_info." in x
    )

    if not host_id_col:
        raise ValueError("Dataframe does not contain an host_id column")

    if "tenant_id" in df.columns:
        tenant_identifier = "tenant_id"
    elif "tenant.id" in df.columns:
        tenant_identifier = "tenant.id"
    else:
        raise ValueError("DataFrame does not contain a valid tenant identifier")

    service = get_service(environment=region)

    tenants_series = df[tenant_identifier].apply(get_tenant_id)
    tenants_list = list(tenants_series.dropna().unique())
    assets_df = pd.DataFrame()

    for tenant in tenants_list:
        host_list = list(df[tenants_series == tenant][host_id_col].dropna().unique())

        with service(tenant_id=tenant):
            for host_ids in chunk_list(host_list, 2000):
                asset_results = assets_from_list(
                    service=service,
                    asset_list=host_ids,
                )
                if asset_results:
                    assets_df = pd.concat(
                        [
                            assets_df,
                            to_dataframe(results=[asdict(x) for x in asset_results])
                            .assign(
                                hostname=lambda x: x.hostnames.apply(
                                    lambda x: x[0].get("hostname", "N/A")
                                )
                            )
                            .add_prefix("asset_info."),
                        ]
                    )
                    continue

    return df.merge(
        assets_df,
        how="left",
        left_on=host_id_col,
        right_on="asset_info.host_id",
    )



"""Custom Subjects API calls."""

from taegis_sdk_python import GraphQLService
from typing import Dict, Any


def lookup_federated_subject(service: GraphQLService) -> Dict[str, Any]:
    """Federated response for expanding the Subjects API.

    Parameters
    ----------
    service : GraphQLService
        Taegis SDK for Python GraphQLService object.

    Returns
    -------
    Dict[str, Any]
        _description_
    """

    output = service.output

    # waiting Union support in GraphQLService
    if not output:
        output = """
            id
            identity {
                __typename
            }
        """

    results = service.subjects.execute_query(endpoint="currentSubject", output=output)
    return results.get("currentSubject", {})



"""Pandas functions for Proccess Lineage Event Dataframes (process only)"""

import logging
import pandas as pd
from typing import List, Optional
from dataclasses import asdict
from taegis_magic.commands.process_trees import process_lineage, process_children


log = logging.getLogger(__name__)


def lookup_lineage(
    df: pd.DataFrame, region: Optional[str] = None, tenant_id: Optional[str] = None
) -> pd.DataFrame:
    """
    For each row in the DataFrame, fetch process lineage using process_correlation_id, host_id, tenant_id, and resource_id.
    Adds a new column 'process_lineage' with the results.

    Parameters
    ----------
    df : pd.DataFrame
        A pandas DataFrame that contains Taegis process events with process_correlation_id, host_id, tenant_id, and resource_id.
    tenant_id: Optional[str]
        Taegis SDK Tenant ID that the asset lookup is for.  Defaults to None.
    region : Optional[str]
        Taegis SDK Region/Environment that the asset lookup is for.  Defaults to US1.

    Returns
    -------
    pd.DataFrame
        Returns a pandas DataFrame with additional asset information columns.

    Raises
    ------
    ValueError
        If the event is not a process event, or if process_correlation_id, host_id, and resource_id does not have columns in the dataframe a value error will be raised.
        If there the dataframe does not contain a valid tenant identifier.
    """

    df = df.copy()
    if df.empty:
        return df

    required_cols = ["host_id", "process_correlation_id", "resource_id"]
    if not all(x in df.columns for x in required_cols):
        raise ValueError(
            "DataFrame must contain host_id, process_correlation_id, and resource_id columns for lineage lookup."
        )

    def get_lineage(row):
        if (
            pd.isna(row.get("host_id"))
            or pd.isna(row.get("process_correlation_id"))
            or pd.isna(row.get("resource_id"))
        ):
            return []

        else:
            result = process_lineage(
                region=region,
                tenant_id=tenant_id,
                host_id=row.get("host_id"),
                process_correlation_id=row.get("process_correlation_id"),
                resource_id=row.get("resource_id"),
            )

        # Check if there are results in the process_lineage call. If not, return empty.
        if hasattr(result, "results") and result.results is not None:
            return result.results
        else:
            return []

    df["process_info.process_lineage"] = df.apply(get_lineage, axis=1)
    df = df.explode("process_info.process_lineage").reset_index(drop=True)
    df["process_info.process_lineage"] = df["process_info.process_lineage"].apply(
        lambda x: {} if pd.isnull(x) else x
    )
    df["process_info.process_lineage.index"] = df["process_info.process_lineage"].apply(
        lambda x: int(x.get("lineage_index", None))
        if x.get("lineage_index", None) is not None
        else None
    )
    return df


def lookup_children(
    df: pd.DataFrame, region: Optional[str] = None, tenant_id: Optional[str] = None
) -> pd.DataFrame:
    """
    For each row in the DataFrame, fetch their children processes using process_correlation_id, host_id, tenant_id, and resource_id.
    Adds a new column 'process_children' with the results.

    Parameters
    ----------
    df : pd.DataFrame
        A pandas DataFrame that contains Taegis process events with process_correlation_id, host_id, tenant_id, and resource_id.
    tenant_id: Optional[str]
        Taegis SDK Tenant ID that the asset lookup is for.  Defaults to None.
    region : Optional[str]
        Taegis SDK Region/Environment that the asset lookup is for.  Defaults to US1.

    Returns
    -------
    pd.DataFrame
        Returns a pandas DataFrame with additional asset information columns.

    Raises
    ------
    ValueError
        If the event is not a process event, or if process_correlation_id, host_id, and resource_id does not have columns in the dataframe a value error will be raised.
    """

    df = df.copy()
    if df.empty:
        return df

    required_cols = ["host_id", "process_correlation_id", "resource_id"]
    if not all(x in df.columns for x in required_cols):
        raise ValueError(
            "DataFrame must contain host_id, process_correlation_id, and resource_id columns for lineage lookup."
        )

    def get_children(row):
        if (
            pd.isna(row.get("host_id"))
            or pd.isna(row.get("process_correlation_id"))
            or pd.isna(row.get("resource_id"))
        ):
            return []

        else:
            result = process_children(
                region=region,
                tenant_id=tenant_id,
                host_id=row.get("host_id"),
                process_correlation_id=row.get("process_correlation_id"),
                resource_id=row.get("resource_id"),
            )

        # Check if there are results in the process_children call. If not, return empty.
        if hasattr(result, "results") and result.results is not None:
            return result.results
        else:
            return []

    df["process_info.process_children"] = df.apply(get_children, axis=1)
    df = df.explode("process_info.process_children").reset_index(drop=True)
    df["process_info.process_children"] = df["process_info.process_children"].apply(
        lambda x: {} if pd.isnull(x) else x
    )
    return df



"""Utility functions for use with Pandas."""

import logging
from typing import List

import pandas as pd

logger = logging.getLogger(__name__)

MAGIC_COLUMN = "taegis_magic.{}"


DEFAULT_COLUMNS = {
    "agentdetection": [
        'host_id',
        'hostname',
        'detection_category',
        'detection_type',
        'image_path',
        'summary',
    ],
    "antivirus": [
        'host_id',
        'action_taken',
        'threat_category',
        'threat_name',
        'user_name',
        'file_path',
    ],
    "apicall": [
        'host_id',
        'hostname',
        'function_called',
        'was_modification_allowed',
        'was_operation_successful',
        'action',
        'commandline',
    ],
    "auth": [
        "source_address",
        "target_user_name",
        "action",
        "auth_system",
        "user_agent",
        "application_name",
    ],
    "cloudaudit": [
        "source_address",
        "user_name",
        "event_type",
        "event_name",
        "mfa_used",
        "user_agent",
    ],
    "detectionfinding": [],
    "dhcp": [
        'host_id',
        'hostname',
        'server_address',
        'client_address',
        'action',
    ],
    "dnsquery": [
        "hostname",
        "os.os",
        "os.arch",
        "processcorrelationid.pid",
        "query_name",
        "query_type",
    ],
    "email": [
        'host_id',
        'direction',
        'status',
        'event_type',
        'from_email_address',
        'subject',
    ],
    "encrypt": [
        'host_id',
        'hostname',
        'source_address',
        'destination_address',
        'tls_version',
    ],
    "filemod": [
        'host_id',
        'hostname',
        'parent_path',
        'process_image_path',
        'action',
        'file_name',
        'process_username',
    ],
    "generic": [
        'host_id',
        'hostname',
        'summary',
    ],
    "http": [
        "source_username",
        "source_address",
        "destination_address",
        "destination_port",
        "user_agent",
        "http_method",
        "response_code",
        "uri_host",
        "uri_path",
        "tx_byte_count",
        "rx_byte_count",
    ],
    "managementevent": [
        'host_id',
        'hostname',
        'type',
        'channel',
        'operation',
        'username',
    ],
    "netflow": [
        "hostname",
        "sensor_type",
        "protocol",
        "source_address",
        "destination_address",
        "destination_port",
        "dns_name",
    ],
    "nids": [
        "source_address",
        "destination_address",
        "action",
        "blocked",
        "enrichSummary",
    ],
    "persistence": [
        'host_id',
        'hostname',
        'category',
        'command.program.path',
        'command.args'
    ],
    "process": [
        "hostname",
        "os.os",
        "os.arch",
        "sensor_type",
        "username",
        "user_is_admin",
        "process_is_admin",
        "image_path",
        "commandline",
        "was_blocked",
    ],
    "processmodule": [
        'host_id',
        'hostname',
        'sensor_action',
        'file.path',
        'module_action',
    ],
    "registry": [
        "host_id",
        "hostname",
        "event_type",
        "path",
        "process_image_path",
        "process_username",
        "value.name",
        "value.data32",
        "value.data64",
    ],
    "scriptblock": [
        "os.os",
        "os.arch",
        "interpreter_name",
        "interpreter_path",
        "script_name",
        "decoded_block_text",
    ],
    "thirdparty": [
        "source_address",
        "user_principal_name",
        "title",
        "ontology",
    ],
    'threadinjection': [
        'host_id',
        'hostname',
        'source_process_name',
        'target_process_name',
        'thread_id',
    ],
}


def get_tenant_id(tenant_id):
    """Coerce tenant ids into common format."""
    if isinstance(tenant_id, int):
        return str(tenant_id)
    elif isinstance(tenant_id, list):
        return str(tenant_id[0])
    elif isinstance(tenant_id, str):
        replacement_chars = ["[", "]", "'", '"']
        for replacement_char in replacement_chars:
            tenant_id = tenant_id.replace(replacement_char, "")
        return tenant_id
    else:
        raise ValueError(f"{tenant_id} is invalid format")


def chunk_list(lst, n):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i : i + n]


def coalesce_columns(df: pd.DataFrame, columns: List[str]) -> pd.Series:
    """Reduce results to the first result in list of columns."""
    from functools import reduce

    return reduce(
        lambda left, right: left.combine_first(right),
        [df[col] for col in columns if col in df.columns],
    )


def return_valid_column(df: pd.DataFrame, column_list: List[str]) -> pd.Series:
    """Takes a dataframe and a list of user defined columns to obtain the first possible valid column.

    Parameters
    ----------
    df : pd.DataFrame
        A pandas dataframe.
    column_list : List[str]
        A list of columns a user defines to pull out the first possible valid column
        from the dataframe that is passed.

    Returns
    -------
    pd.Series
        Returns a pandas Series of the first column it was able to obtain from the dataframe passed in.

    Raises
    ------
    ValueError
        Returns if there are no valid columns found within the dataframe.
    """
    valid_columns = []

    if not column_list or not isinstance(column_list, list):
        raise ValueError("Provided column list is either blank or not a list.")

    for col in column_list:
        if col in df.columns:
            valid_columns.append(col)
            logger.debug(f"Found identifier column: {col}")

    if not valid_columns:
        raise ValueError(
            f"DataFrame does not contain any columns from supplied list: {column_list}"
        )

    return coalesce_columns(df, valid_columns)


def drop_duplicates_by_hashables(df: pd.DataFrame) -> pd.DataFrame:
    """
    Drop duplicates from a Pandas DataFrame based upon it's hashable columns.

    Parameters
    ----------
    df : pd.DataFrame
        Pandas DataFrame

    Returns
    -------
    pd.DataFrame
        Pandas DataFrame
    """
    hashable_columns = []
    for column in df.columns:
        try:
            df.drop_duplicates([df.columns[0], column])
        except TypeError:
            # log.debug(f"{column} is non-hashable skipping...")
            continue

        hashable_columns.append(column)

    return df.copy().drop_duplicates(hashable_columns)


def groupby(df: pd.DataFrame, columns: List[str]) -> pd.DataFrame:
    """Validate column list against DataFrame and return a standardized Pandas DataFrame goupby operation.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame to group by.
    columns : List[str]
        List of columns to group by.

    Returns
    -------
    pd.DataFrame
        DataFrame with grouped counts.
    """
    for column in columns.copy():
        if not column in df.columns:
            logger.error(f"Column {column} not found in dataframe")
            columns.remove(column)
    return df[columns].astype(str).groupby(columns, dropna=False).size().reset_index(name="count")


def default_schema_columns(schema: str) -> List[str]:
    """Return default Magic columns for a given schema.

    Parameters
    ----------
    schema : str
        Schema name.

    Returns
    -------
    List[str]
        List of default columns.
    """
    return DEFAULT_COLUMNS.get(schema.lower(), [])


"""Pandas function for Audits DataFrames."""

import difflib
import logging

import pandas as pd
import warnings

warnings.simplefilter(action="ignore", category=pd.errors.PerformanceWarning)

logger = logging.getLogger(__name__)


def get_diffs(df: pd.DataFrame) -> pd.DataFrame:
    """Pull out differences in before_state and after_state.

    Parameters
    ----------
    df : pd.DataFrame
        Audits DataFrame

    Returns
    -------
    pd.DataFrame
        Audits with before_state and after_state diff side by side.

    Example
    -------
    Example::

        %taegis audits search --application investigations --assign search_audits

        search_audits = search_audits.pipe(get_diffs)

    """
    if df.empty:
        return df

    df = df.copy()

    def get_diff(row, col_name: str):
        differ = difflib.Differ()

        if isinstance(row[f"before_state.{col_name}"], list):
            before_state = [str(item) for item in row[f"before_state.{col_name}"]]
        else:
            before_state = str(row[f"before_state.{col_name}"]).splitlines(
                keepends=True
            )

        if isinstance(row[f"after_state.{col_name}"], list):
            after_state = [str(item) for item in row[f"after_state.{col_name}"]]
        else:
            after_state = str(row[f"after_state.{col_name}"]).splitlines(keepends=True)

        diff = list(
            differ.compare(
                before_state,
                after_state,
            )
        )
        return [
            item
            for item in diff
            if item.startswith("-") or item.startswith("+") or item.startswith("?")
        ]

    before_state_columns = {
        column.replace("before_state.", "")
        for column in df.columns
        if column.lower().startswith("before_state")
    }
    after_state_columns = {
        column.replace("after_state.", "")
        for column in df.columns
        if column.lower().startswith("after_state")
    }
    if not before_state_columns:
        raise ValueError("DataFrame does not have a before_state")
    if not after_state_columns:
        raise ValueError("DataFrame does not have an after_state")

    diff_columns = list(before_state_columns.intersection(after_state_columns))

    if not diff_columns:
        raise ValueError(
            "DataFrame does not have any intersecting before_state to after_state columns"
        )

    for col in diff_columns:
        df[f"before_state.{col}"] = df[f"before_state.{col}"].fillna("")

        df[f"after_state.{col}"] = df[f"after_state.{col}"].fillna("")

    for column in diff_columns:
        column_name = f"taegis_magic.diff.{column}"
        if column_name not in df.columns:
            logger.debug("Getting difference on %s...", column_name)
            df[column_name] = df.apply(get_diff, args=(column,), axis=1)
        else:
            logger.debug("%s found, moving to next column...", column_name)

    return df



import inspect
import logging
from typing import Optional, List

import typer
from typing_extensions import Annotated

from dataclasses import asdict, dataclass, field
from taegis_magic.core.log import tracing
from taegis_magic.core.normalizer import TaegisResults, TaegisResult
from taegis_magic.core.service import get_service

app = typer.Typer(help="Taegis Process Trees Commands.")

log = logging.getLogger(__name__)

from dataclasses import asdict, dataclass
from taegis_magic.core.normalizer import TaegisResultsNormalizer
from taegis_sdk_python.services.process_trees.types import ProcessLineage


@dataclass
class ProcessLineageNormalizer(TaegisResultsNormalizer):
    """Normalizer for ProcessLineage results, always returns a single dataclass with lineage entries flattened into rows."""

    raw_results: ProcessLineage = field(default_factory=lambda: ProcessLineage())

    @property
    def results(self):
        rows = []
        for result in (
            self.raw_results
            if isinstance(self.raw_results, list)
            else [self.raw_results]
        ):
            if hasattr(result, "__dataclass_fields__"):
                result_dict = asdict(result)
            else:
                result_dict = result
            lineage = result_dict.get("lineage", [])
            if lineage is None:
                lineage = []
            for idx, entry in enumerate(lineage):
                row = {"lineage_index": idx}
                # if entry is not a dict, convert to dict and append to row.
                if hasattr(entry, "__dataclass_fields__"):
                    row.update(asdict(entry))
                elif isinstance(entry, dict):
                    row.update(entry)
                rows.append(row)
        return rows


@app.command()
@tracing
def process_lineage(
    region: Annotated[Optional[str], typer.Option(help="Taegis region")] = None,
    tenant_id: Annotated[Optional[str], typer.Option(help="Taegis tenant id")] = None,
    host_id: Annotated[Optional[str], typer.Option(help="Taegis host id")] = None,
    process_correlation_id: Annotated[
        Optional[str], typer.Option(help="Taegis Event process correlation id")
    ] = None,
    resource_id: Annotated[
        Optional[str], typer.Option(help="Taegis Event resource ID")
    ] = None,
):
    """Get process lineage for a given region & tenant, based on the resource_id, host_id, and process_correlation_id."""
    service = get_service(environment=region, tenant_id=tenant_id)

    results = service.process_trees.query.process_lineage(
        host_id=host_id,
        process_correlation_id=process_correlation_id,
        tenant_id=tenant_id,
        resource_id=resource_id,
    )
    # Process Lineage does not have any pagination, so it always returns a single dataclass that needs to be normalized differently to reflect on the dataframe.

    normalized_results = ProcessLineageNormalizer(
        raw_results=results,
        service="process_lineage",
        tenant_id=service.tenant_id,
        region=service.environment,
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


@app.command()
@tracing
def process_children(
    region: Annotated[Optional[str], typer.Option(help="Taegis region")] = None,
    tenant_id: Annotated[Optional[str], typer.Option(help="Taegis tenant id")] = None,
    host_id: Annotated[str, typer.Option(help="Taegis host id")] = None,
    process_correlation_id: Annotated[
        str, typer.Option(help="Taegis Event process correlation id")
    ] = None,
    resource_id: Annotated[str, typer.Option(help="Taegis Event resource ID")] = None,
):
    """Get process children for a given region & tenant, based on the resource_id, host_id, and process_correlation_id."""
    service = get_service(environment=region, tenant_id=tenant_id)
    all_results = []
    next_token = None

    # Process children has pagination, so we need to loop through all next_token calls until there are no more results.
    while True:
        results = service.process_trees.query.process_children(
            host_id=host_id,
            process_correlation_id=process_correlation_id,
            tenant_id=tenant_id,
            resource_id=resource_id,
            next_token=next_token,
        )

        all_results.extend(results.process_list)
        next_token = results.next_token
        if not next_token:
            break

    normalized_results = TaegisResults(
        raw_results=all_results,
        service="process_children",
        tenant_id=service.tenant_id,
        region=service.environment,
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


@app.command()
@tracing
def process_parent(
    region: Annotated[Optional[str], typer.Option(help="Taegis region")] = None,
    tenant_id: Annotated[Optional[str], typer.Option(help="Taegis tenant id")] = None,
    host_id: Annotated[str, typer.Option(help="Taegis host id")] = None,
    parent_pcid: Annotated[
        str, typer.Option(help="Taegis Event process correlation id")
    ] = None,
    resource_id: Annotated[str, typer.Option(help="Taegis Event resource ID")] = None,
):
    """Gets the processes' parent based on the parent_correlation_id for a given region & tenant, based on the resource_id, host_id, and parent_process_correlation_id."""
    service = get_service(environment=region, tenant_id=tenant_id)
    results = service.process_trees.query.process_parent(
        host_id=host_id,
        parent_process_correlation_id=parent_pcid,
        tenant_id=tenant_id,
        resource_id=resource_id,
    )

    # There is always only one parent process returned, so we can normalize it directly with TaegisResult.
    normalized_results = TaegisResult(
        raw_results=results,
        service="process_parent",
        tenant_id=service.tenant_id,
        region=service.environment,
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results



"""Taegis Magic tenants commands."""

import inspect
import json
import logging
from dataclasses import asdict, dataclass, field
from typing import Any, Dict, List, Tuple, Optional

import typer
from dataclasses_json import config, dataclass_json
from taegis_magic.core.log import tracing
from taegis_magic.core.normalizer import TaegisResultsNormalizer

from taegis_magic.core.service import get_service
from taegis_sdk_python.services.tenants.types import (
    OrderDir,
    TenantEnvironmentFilter,
    TenantLabelFilter,
    TenantOrderField,
    TenantResults,
    TenantsQuery,
    TimeFilter,
)
from typing_extensions import Annotated
from click.exceptions import BadOptionUsage

log = logging.getLogger(__name__)

app = typer.Typer(help="Taegis Tenant Commands.")


@dataclass_json
@dataclass
class TaegisTenantsResultsNormalizer(TaegisResultsNormalizer):
    """Taegis TenantResults normalizer."""

    raw_results: List[TenantResults] = field(default_factory=lambda: [TenantResults()])

    @property
    def results(self):
        return [
            asdict(tenant) for result in self.raw_results for tenant in result.results
        ]

    @property
    def total_results(self) -> int:
        return int(self.raw_results[0].total_count) if self.raw_results else -1

    @property
    def results_returned(self) -> int:
        """Number of results returned from service."""
        return len(self.results)


@app.command()
@tracing
def search(
    filter_by_name: Annotated[
        Optional[str], typer.Option(help="(supports wildcard %) or tenant id")
    ] = None,
    filter_by_tenant: Annotated[Optional[List[str]], typer.Option()] = None,
    filter_by_region: Annotated[Optional[str], typer.Option()] = None,
    filter_by_service: Annotated[Optional[List[str]], typer.Option()] = None,
    filter_by_label_name: Annotated[Optional[str], typer.Option()] = None,
    filter_by_label_value: Annotated[
        Optional[str], typer.Option(help="--filter-by-label-name is required for use")
    ] = None,
    filter_by_partner_subscription: Annotated[
        Optional[List[str]], typer.Option()
    ] = None,
    filter_by_requested_service: Annotated[Optional[List[str]], typer.Option()] = None,
    filter_by_created_start_time: Annotated[
        Optional[str],
        typer.Option(help="YYYY-MM-DDTHH:MM:SSZ"),
    ] = None,
    filter_by_created_end_time: Annotated[
        Optional[str],
        typer.Option(help="YYYY-MM-DDTHH:MM:SSZ"),
    ] = None,
    filter_by_modified_start_time: Annotated[
        Optional[str],
        typer.Option(help="YYYY-MM-DDTHH:MM:SSZ"),
    ] = None,
    filter_by_modified_end_time: Annotated[
        Optional[str],
        typer.Option(help="YYYY-MM-DDTHH:MM:SSZ"),
    ] = None,
    filter_by_tenant_hierarchy: Annotated[Optional[List[str]], typer.Option()] = None,
    filter_by_enabled: Annotated[Optional[bool], typer.Option()] = None,
    sort_by_field: Annotated[TenantOrderField, typer.Option()] = TenantOrderField.ID,
    sort_order: Annotated[OrderDir, typer.Option()] = OrderDir.ASC,
    tenant: Annotated[Optional[str], typer.Option()] = None,
    region: Annotated[Optional[str], typer.Option()] = None,
):
    """Search Taegis tenants."""
    if filter_by_label_value and not filter_by_label_name:
        raise BadOptionUsage(
            ..., "--filter-by-label-value requires --filter-by-label-name to be set..."
        )

    service = get_service(environment=region, tenant_id=tenant)

    max_results = 1000
    page_number = 1

    log.info(f"Polling page: {page_number}")

    result = service.tenants.query.tenants(
        TenantsQuery(
            max_results=max_results,
            page_num=page_number,
            name=filter_by_name,
            ids=filter_by_tenant,
            for_hierarchies=filter_by_tenant_hierarchy,
            with_partner_subscriptions=filter_by_partner_subscription,
            with_requested_services=filter_by_requested_service,
            label_filter=(
                TenantLabelFilter(
                    label_name=filter_by_label_name, label_value=filter_by_label_value
                )
                if filter_by_label_name
                else None
            ),
            environment_filter=(
                TenantEnvironmentFilter(name=filter_by_region, enabled=True)
                if filter_by_region
                else None
            ),
            created_time_filter=(
                TimeFilter(
                    start_time=filter_by_created_start_time,
                    end_time=filter_by_created_end_time,
                )
                if (filter_by_created_start_time or filter_by_created_end_time)
                else None
            ),
            modified_time_filter=(
                TimeFilter(
                    start_time=filter_by_modified_start_time,
                    end_time=filter_by_modified_end_time,
                )
                if (filter_by_modified_start_time or filter_by_modified_end_time)
                else None
            ),
            with_services=filter_by_service,
            order_by=sort_by_field,
            order_dir=sort_order,
            enabled_in_production=filter_by_enabled,
        )
    )

    results = [result]

    while result.has_more:
        page_number += 1
        log.info(f"Polling page: {page_number}")

        result = service.tenants.query.tenants(
            TenantsQuery(
                max_results=max_results,
                page_num=page_number,
                name=filter_by_name,
                ids=filter_by_tenant,
                for_hierarchies=filter_by_tenant_hierarchy,
                with_partner_subscriptions=filter_by_partner_subscription,
                with_requested_services=filter_by_requested_service,
                label_filter=(
                    TenantLabelFilter(
                        label_name=filter_by_label_name,
                        label_value=filter_by_label_value,
                    )
                    if filter_by_label_name
                    else None
                ),
                environment_filter=(
                    TenantEnvironmentFilter(name=filter_by_region, enabled=True)
                    if filter_by_region
                    else None
                ),
                created_time_filter=(
                    TimeFilter(
                        start_time=filter_by_created_start_time,
                        end_time=filter_by_created_end_time,
                    )
                    if (filter_by_created_start_time or filter_by_created_end_time)
                    else None
                ),
                modified_time_filter=(
                    TimeFilter(
                        start_time=filter_by_modified_start_time,
                        end_time=filter_by_modified_end_time,
                    )
                    if (filter_by_modified_start_time or filter_by_modified_end_time)
                    else None
                ),
                with_services=filter_by_service,
                order_by=sort_by_field,
                order_dir=sort_order,
                enabled_in_production=filter_by_enabled,
            )
        )

        results.append(result)

    normalized_results = TaegisTenantsResultsNormalizer(
        raw_results=results,
        service="tenants",
        tenant_id=service.tenant_id,
        region=service.environment,
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


if __name__ == "__main__":
    rv = app(standalone_mode=False)
    print(json.dumps(rv.results))



"""Taegis Magic clients commands."""

import logging
from dataclasses import field, asdict
from enum import Enum
from pprint import pprint
from typing import List, Optional, Union
import inspect
from typing_extensions import Annotated

import typer
from taegis_magic.core.log import tracing
from taegis_magic.core.normalizer import TaegisResultsNormalizer

from taegis_magic.core.service import get_service
from taegis_sdk_python.services.clients.types import (
    Client,
    NewClient,
    ClientRoleAssignmentInput,
)

log = logging.getLogger(__name__)

app = typer.Typer(help="Taegis Clients Commands.")
rotate_app = typer.Typer(help="Taegis Clients Rotate Commands.")
role_app = typer.Typer(help="Taegis Clients Role Commands.")
app.add_typer(rotate_app, name="rotate")
app.add_typer(role_app, name="role")


class Roles(str, Enum):
    """Role choices for client commands."""

    administrator = "administrator"
    analyst = "analyst"
    responder = "responder"
    auditor = "auditor"


ROLE_MAP = {
    "administrator": "ba0fdcbd-e87d-4bdd-ae7d-ca6118b25068",
    "analyst": "a4903f9f-465b-478f-a24e-82fa2e129d2e",
    "responder": "a72dace7-4536-4dbc-947d-015a8eb65f4d",
    "auditor": "ace1cae4-59fd-4fd1-9500-40077dc529a7",
}


class TaegisClientResultsNormalizer(TaegisResultsNormalizer):
    """Taegis Client Results normalizer."""

    raw_results: Union[Client, NewClient] = field(default_factory=lambda: Client())

    @property
    def results(self):
        return [asdict(self.raw_results)]


class TaegisClientsResultsNormalizer(TaegisResultsNormalizer):
    """Taegis Clients Results normalizer."""

    raw_results: List[Client] = field(default_factory=list)

    @property
    def results(self):
        return [asdict(result) for result in self.raw_results]


@app.command()
@tracing
def create(
    app_name: str,
    tenant: Optional[str] = None,
    region: Optional[str] = None,
    role: Optional[List[Roles]] = None,
):
    """Create an OAuth2 client_id and client_secret for non-interactive login."""
    service = get_service(environment=region, tenant_id=tenant)
    roles = [ROLE_MAP[r.value] for r in role or [Roles.analyst]]

    results = service.clients.mutation.create_client(name=app_name, roles=roles)

    normalized_results = TaegisClientResultsNormalizer(
        raw_results=results,
        tenant_id=service.tenant_id,
        region=service.environment,
        service="clients",
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


@app.command()
@tracing
def remove(
    client_id: Annotated[str, typer.Argument(help="client UUID or client_id")],
    region: Optional[str] = None,
    tenant: Optional[str] = None,
):
    """Delete an OAuth2 client_id."""
    service = get_service(environment=region, tenant_id=tenant)

    results = service.clients.mutation.delete_client(client_id)

    normalized_results = TaegisClientResultsNormalizer(
        raw_results=results,
        tenant_id=service.tenant_id,
        region=service.environment,
        service="clients",
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


@app.command()
@tracing
def search(
    name: Optional[str] = None,
    client_id: Optional[List[str]] = None,
    role: Optional[List[Roles]] = None,
    region: Optional[str] = None,
    tenant: Optional[str] = None,
):
    """Search OAuth2 clients."""
    service = get_service(environment=region, tenant_id=tenant)
    if role:
        roles = [ROLE_MAP[r.value] for r in role]
    else:
        roles = None

    results = service.clients.query.clients(
        name=name, client_ids=client_id, role_ids=roles
    )

    normalized_results = TaegisClientsResultsNormalizer(
        raw_results=results,
        tenant_id=service.tenant_id,
        region=service.environment,
        service="clients",
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


@rotate_app.command(name="secret")
@tracing
def rotate_secret(
    client_id: str,
    region: Optional[str] = None,
    tenant: Optional[str] = None,
):
    """Rotate secret value for an OAuth2 client."""
    service = get_service(environment=region, tenant_id=tenant)

    results = service.clients.mutation.rotate_client_secret(client_id)

    normalized_results = TaegisClientResultsNormalizer(
        raw_results=results,
        tenant_id=service.tenant_id,
        region=service.environment,
        service="clients",
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


@role_app.command(name="append")
@tracing
def role_append(
    client_id: str,
    role: Annotated[List[Roles], typer.Option()],
    tenant: Annotated[str, typer.Option()],
    region: Optional[str] = None,
):
    """Add/Append a role to an OAuth2 client."""
    service = get_service(environment=region, tenant_id=tenant)
    roles = [ROLE_MAP[r.value] for r in role]

    results = service.clients.mutation.append_client_roles(
        client_id,
        roles=[
            ClientRoleAssignmentInput(tenant_id=tenant, role_id=role) for role in roles
        ],
    )

    normalized_results = TaegisClientResultsNormalizer(
        raw_results=results,
        tenant_id=service.tenant_id,
        region=service.environment,
        service="clients",
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


@role_app.command(name="remove")
@tracing
def role_remove(
    client_id: str,
    role: Annotated[List[Roles], typer.Option()],
    tenant: Annotated[str, typer.Option()],
    region: Optional[str] = None,
):
    """Remove a role for an OAuth2 client."""
    service = get_service(environment=region, tenant_id=tenant)
    roles = [ROLE_MAP[r.value] for r in role]

    results = service.clients.mutation.remove_client_roles(
        client_id,
        roles=roles,
    )

    normalized_results = TaegisClientResultsNormalizer(
        raw_results=results,
        tenant_id=service.tenant_id,
        region=service.environment,
        service="clients",
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


if __name__ == "__main__":
    rv = app(standalone_mode=False)
    pprint(rv)



"""Taegis Magic threat commands."""

import logging
import inspect
from dataclasses import asdict, dataclass, field
from typing import Any, Dict, List, Optional
from urllib.parse import quote

import typer
from dataclasses_json import dataclass_json
from typing_extensions import Annotated

from taegis_magic.core.log import tracing
from taegis_magic.core.normalizer import TaegisResultsNormalizer, TaegisResults
from taegis_magic.core.service import get_service
from taegis_sdk_python.services.threat.types import ThreatParentType, ThreatPublication

log = logging.getLogger(__name__)

app = typer.Typer(help="Taegis Threat Intelligence Commands.")
publications_app = typer.Typer(help="Search CTU Threat Publications.")
app.add_typer(publications_app, name="publications")


@dataclass_json
@dataclass
class ThreatPublicationsNormalizer(TaegisResultsNormalizer):
    """Threat Publications Normalizer."""

    raw_results: List[ThreatPublication] = field(default_factory=list)

    @property
    def results(self) -> List[Dict[str, Any]]:
        """Results."""
        service = get_service(environment=self.region)

        # a reference to the threat intel article doesn't exist in the
        # data, so we will generate one based on the region used
        return [
            {
                **asdict(pub),
                "taegis_magic.reference": (
                    f"{service.core.sync_url.replace('api.', '')}/threat-intelligence-publications?"
                    f"headerText=Threat%20Intelligence%20Reports&id={quote(pub.id or '')}"
                ),
            }
            for pub in self.raw_results
        ]


@publications_app.command("latest")
@tracing
def publications_latest(
    size: int,
    region: Optional[str] = None,
):
    """
    Retrieve the latest CTU threat publications.
    """
    service = get_service(environment=region)

    publications = service.threat.query.threat_latest_publications(from_=0, size=size)

    results = ThreatPublicationsNormalizer(
        service="threat",
        tenant_id=service.tenant_id,
        region=service.environment,
        raw_results=publications,
        arguments={
            "size": size,
            "region": region,
        },
    )

    return results


@publications_app.command("search")
@tracing
def publications_search(
    term: str,
    region: Optional[str] = None,
):
    """
    Search the CTU threat publications.
    """
    service = get_service(environment=region)

    publications = service.threat.query.threat_publications(text=term)

    results = ThreatPublicationsNormalizer(
        service="threat",
        tenant_id=service.tenant_id,
        region=service.environment,
        raw_results=publications,
        arguments={
            "term": term,
            "region": region,
        },
    )

    return results


@app.command()
@tracing
def watchlist(
    type_: Annotated[ThreatParentType, typer.Option("--type")],
    tenant: Annotated[Optional[str], typer.Option()] = None,
    region: Annotated[Optional[str], typer.Option()] = None,
):
    """
    Retrieve CTU watchlist by indicator type.
    """
    service = get_service(tenant_id=tenant, environment=region)
    results = service.threat.query.threat_watchlist(type_)

    normalized_results = TaegisResults(
        service="threat",
        tenant_id=service.tenant_id,
        region=service.environment,
        raw_results=results,
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results



"""Taegis Magic alerts commands."""

import logging
from dataclasses import asdict, dataclass, field
from pprint import pprint
from typing import Any, Dict, List, Optional

import typer
from dataclasses_json import config, dataclass_json
from taegis_magic.commands.configure import QUERIES_SECTION
from taegis_magic.core.log import tracing
from taegis_magic.core.normalizer import TaegisResultsNormalizer, TaegisResults

from taegis_magic.core.service import get_service
from taegis_sdk_python import (
    GraphQLNoRowsInResultSetError,
    GraphQLService,
    build_output_string,
    prepare_input,
)
from taegis_sdk_python.config import get_config
from taegis_sdk_python.services.alerts.types import (
    Alert2,
    AlertsList,
    AlertsResponse,
    AuxiliaryEvent,
    PollRequestInput,
    SearchRequestInput,
)
from taegis_sdk_python.services.sharelinks.types import (
    ExtraParamCreateInput,
    ShareLinkCreateInput,
)
from taegis_magic.commands.utils.investigations import insert_search_query
from typing_extensions import Annotated


log = logging.getLogger(__name__)

CONFIG = get_config()
if not CONFIG.has_section(QUERIES_SECTION):
    CONFIG.add_section(QUERIES_SECTION)

app = typer.Typer(help="Taegis Alerts Commands.")


@dataclass_json
@dataclass
class AlertsResultsNormalizer(TaegisResultsNormalizer):
    """Taegis Alerts Normalizer."""

    raw_results: List[AlertsResponse] = field(default_factory=list)
    query: str = ""
    is_saved: bool = False
    _query_id: Optional[str] = None
    _shareable_url: Optional[str] = None

    def _repr_markdown_(self):
        """Represent as markdown."""
        return self._display_template(
            "taegis_search_results.md.jinja"
        )  # pragma: no test

    @property
    def results(self) -> List[Dict[str, Any]]:
        log.debug("Calling AlertsResultsNormalizer.results...")
        if not self.raw_results:
            return []

        if self.raw_results[0].alerts.group_by:
            return self.aggregate

        return [
            asdict(alert) for result in self.raw_results for alert in result.alerts.list
        ]

    @property
    def status(self) -> str:
        """Alerts Service results status."""
        log.debug("Calling AlertsResultsNormalizer.status...")
        return self.raw_results[0].status.value if self.raw_results else "Error"

    @property
    def total_results(self) -> int:
        log.debug("Calling AlertsResultsNormalizer.num_total...")
        # This signifies an error
        if not self.raw_results:
            return -1

        # This signifies a protobuf error, where 0 results doesn't return a total_results value
        return (
            self.raw_results[0].alerts.total_results
            if self.raw_results[0].alerts.total_results
            else 0
        )

    @property
    def results_returned(self) -> int:
        """Query results total number of results returned.

        Returns
        -------
        str
            Returns number of results.
        """
        log.debug("Calling AlertsResultsNormalizer.total_returned...")
        return (
            sum([len(result.alerts.list) for result in self.raw_results])
            if self.raw_results
            else -1
        )

    @property
    def aggregate(self) -> List[Dict[str, Any]]:
        """
        Taegis Alerts aggregate results parser.

        Returns
        -------
        List[Dict[str, Any]]
            List of aggregate results
        """
        log.debug("Calling AlertsResultsNormalizer.aggregate...")
        if not self.raw_results[0].alerts.group_by:
            return []

        aggs = []
        for response in self.raw_results[0].alerts.group_by:
            agg = {key.key: key.value for key in response.keys}
            agg["count"] = response.value
            aggs.append(agg)

        return aggs

    @property
    def query_identifier(self) -> Optional[str]:
        """Alerts Service Query Identifier."""
        log.debug("Calling AlertsResultsNormalizer.query_identifier...")
        if not self.raw_results:
            return None

        if self.raw_results[0].query_id:
            return self.raw_results[0].query_id

        return None

    @property
    def shareable_url(self) -> str:
        """Alerts Service Sharelinks URL."""
        log.debug("Calling AlertsResultsNormalizer.shareable_url...")
        if self._shareable_url:
            return self._shareable_url

        if not self.raw_results:
            return "Unable to create shareable link"

        if self.aggregate:
            return "Unable to create shareable link"

        if not self.query_identifier:
            return "Unable to create shareable link"

        service = get_service(environment=self.region, tenant_id=self.tenant_id)

        result = service.sharelinks.mutation.create_share_link(
            ShareLinkCreateInput(
                link_ref=self.query_identifier,
                link_target="cql",
                link_type="queryId",
                tenant_id=self.tenant_id,
                extra_parameters=[
                    ExtraParamCreateInput(key="sourceType", value="alert"),
                ],
            )
        )

        self._shareable_url = (
            f'{service.core.sync_url.replace("api.", "")}/share/{result.id}'
        )
        return self._shareable_url


@dataclass_json
@dataclass(order=True, eq=True, frozen=True)
class CustomAuxiliaryEvent(AuxiliaryEvent):
    """My Custom Auxiliary Event - Extends Auxiliary Event with event_data
    to take advantage of GQL federated services.
    """

    event_data: Optional[Dict[str, Any]] = field(
        default=None, metadata=config(field_name="event_data")
    )


@dataclass_json
@dataclass(order=True, eq=True, frozen=True)
class CustomAlert2(Alert2):
    """My Custom Alert2."""

    event_ids: Optional[List[CustomAuxiliaryEvent]] = field(
        default=None, metadata=config(field_name="event_ids")
    )


@dataclass_json
@dataclass(order=True, eq=True, frozen=True)
class CustomAlertsList(AlertsList):
    """My Custom AlertsList."""

    list: Optional[List[CustomAlert2]] = field(
        default=None, metadata=config(field_name="list")
    )


@dataclass_json
@dataclass(order=True, eq=True, frozen=True)
class CustomAlertsResponse(AlertsResponse):
    """My Custom AlertsResponse."""

    alerts: Optional[CustomAlertsList] = field(
        default=None, metadata=config(field_name="alerts")
    )


@tracing
def alerts_service_search_with_events(
    service: GraphQLService, in_: SearchRequestInput
) -> CustomAlertsResponse:
    """Query Taegis Alerts with corresponding Events attached."""
    endpoint = "alertsServiceSearch"
    result = service.alerts.execute_query(
        endpoint=endpoint,
        variables={
            "in": prepare_input(in_),
        },
        output=build_output_string(CustomAlertsResponse),
    )
    if result is not None:
        return CustomAlertsResponse.from_dict(result.get(endpoint))
    raise GraphQLNoRowsInResultSetError("for query alertsServiceSearch")


@tracing
def alerts_service_poll_with_events(
    service: GraphQLService, in_: PollRequestInput
) -> CustomAlertsResponse:
    endpoint = "alertsServicePoll"
    result = service.alerts.execute_query(
        endpoint=endpoint,
        variables={
            "in": prepare_input(in_),
        },
        output=build_output_string(CustomAlertsResponse),
    )
    if result is not None:
        return CustomAlertsResponse.from_dict(result.get(endpoint))
    raise GraphQLNoRowsInResultSetError("for custom query alertsServicePoll")


@app.command()
@tracing
def search(
    cell: Optional[str] = None,
    region: Optional[str] = None,
    tenant: Optional[str] = None,
    limit: int = 10000,
    graphql_output: Optional[str] = None,
    track: Annotated[bool, typer.Option()] = CONFIG[QUERIES_SECTION].getboolean(
        "track", fallback=False
    ),
    database: Annotated[str, typer.Option()] = ":memory:",
) -> Optional[AlertsResultsNormalizer]:
    """
    Search Taegis Alerts service.
    """
    service = get_service(environment=region, tenant_id=tenant)
    if not cell:
        cell = ""

    if "aggregate" in cell:
        limit = 1

    with service(output=graphql_output):
        result = alerts_service_search_with_events(
            service,
            SearchRequestInput(
                cql_query=cell,
                offset=0,
                limit=limit,
                metadata={
                    "callerName": CONFIG[QUERIES_SECTION].get(
                        "callername", fallback="Taegis Magic"
                    ),
                },
            ),
        )

    poll_responses = [result]
    search_id = result.search_id
    total_parts = result.alerts.total_parts

    if search_id:
        for part in range(2, total_parts + 1):
            response = None
            try:
                log.debug(f"Submitting page {part}...")
                with service(output=graphql_output):
                    response = alerts_service_poll_with_events(
                        service,
                        PollRequestInput(
                            search_id=search_id,
                            part_id=part,
                        ),
                    )
            except Exception as exc:
                log.error(
                    f"Cannot retrieve results for search_id:{search_id}:{part}::{exc}"
                )
                if "not found" in str(exc):
                    break

            if isinstance(response, AlertsResponse) and response.alerts is not None:
                poll_responses.append(response)
                # CX-92571 work around
                if sum(len(response.alerts.list) for response in poll_responses) >= int(
                    limit
                ):
                    break

    results = AlertsResultsNormalizer(
        raw_results=poll_responses,
        service="alerts",
        tenant_id=service.tenant_id,
        region=service.environment,
        query=cell,
        arguments={
            "cell": cell,
            "region": service.environment,
            "tenant": service.tenant_id,
            "limit": limit,
            "graphql_output": graphql_output,
        },
    )

    if track:
        insert_search_query(database, results)

    return results


@app.command()
@tracing
def history(
    id_: Annotated[str, typer.Option("--id", help="Alert ID")],
    tenant: Annotated[Optional[str], typer.Option(help="Taegis Tenant ID")] = None,
    region: Annotated[Optional[str], typer.Option(help="Taegis Region")] = None,
) -> TaegisResults:
    """Retrieve Alert history by ID."""
    service = get_service(tenant_id=tenant, environment=region)

    results = service.alerts_history.query.alert_history_by_id(id_=id_)

    return TaegisResults(
        raw_results=results,
        service="alerts_history",
        tenant_id=service.tenant_id,
        region=service.environment,
    )


if __name__ == "__main__":
    rv = app(standalone_mode=False)
    pprint(rv)



"""Taegis Magic audits commands."""

import inspect
import logging
from dataclasses import asdict, dataclass, field
from pprint import pprint
from typing import Any, Dict, List, Optional

import typer
from dataclasses_json import config, dataclass_json
from taegis_magic.core.log import tracing
from taegis_magic.core.normalizer import TaegisResultsNormalizer
from taegis_magic.core.service import get_service
from taegis_sdk_python.services.audits.types import (
    AllAuditsInput,
    Audit,
    AuditEventEnum,
    AuditSearchInput,
    SortBy,
    SortOrder,
    AuditResult,
    AuditEventResult,
)
from typing_extensions import Annotated

log = logging.getLogger(__name__)

app = typer.Typer(help="Taegis Audit Commands.")


@dataclass_json
@dataclass
class TaegisAuditNormalizer(TaegisResultsNormalizer):
    """Signle Taegis Audit Results Normalizer."""

    raw_results: Audit = field(default_factory=lambda: Audit())

    @property
    def results(self):
        return [asdict(self.raw_results)]


@dataclass_json
@dataclass
class TaegisAuditResultNormalizer(TaegisResultsNormalizer):
    """Taegis AuditResult Results Normalizer."""

    raw_results: AuditResult = field(default_factory=lambda: AuditResult())

    @property
    def results(self):
        return [asdict(audit) for audit in self.raw_results.audits or []]

    @property
    def status(self) -> str:
        """Status of query.

        Returns
        -------
        str
            Status message of query.
        """
        return "SUCCESS" if self.raw_results else "ERROR"

    @property
    def total_results(self) -> int:
        # This signifies an error
        if not self.raw_results:
            return -1

        # This signifies a protobuf error, where 0 results doesn't return a total_results value
        return self.raw_results.total_results or -1

    @property
    def results_returned(self) -> int:
        """Query results total number of results returned.

        Returns
        -------
        str
            Returns number of results.
        """
        return len(self.raw_results.audits) if self.raw_results else -1


@dataclass_json
@dataclass
class TaegisAuditEventResultNormalizer(TaegisResultsNormalizer):
    """Taegis AuditResult Results Normalizer."""

    raw_results: AuditEventResult = field(default_factory=lambda: AuditEventResult())

    @property
    def results(self):
        return [asdict(event) for event in self.raw_results.audit_events or []]

    @property
    def status(self) -> str:
        """Status of query.

        Returns
        -------
        str
            Status message of query.
        """
        return "SUCCESS" if self.raw_results else "ERROR"

    @property
    def total_results(self) -> int:
        # This signifies an error
        if not self.raw_results:
            return -1

        # This signifies a protobuf error, where 0 results doesn't return a total_results value
        return self.raw_results.total_events or -1

    @property
    def results_returned(self) -> int:
        """Query results total number of results returned.

        Returns
        -------
        str
            Returns number of results.
        """
        return len(self.raw_results.audit_events) if self.raw_results else -1


@app.command(name="audit")
@tracing
def audits_audit(
    audit_id: Annotated[str, typer.Option()],
    tenant: Optional[str] = None,
    region: Optional[str] = None,
):
    """Taegis audits get a single audit record."""
    service = get_service(environment=region, tenant_id=tenant)

    results = service.audits.query.audit(audit_id=audit_id)

    normalized_results = TaegisAuditNormalizer(
        raw_results=results,
        service="audits",
        tenant_id=service.tenant_id,
        region=service.environment,
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


@app.command(name="all")
@tracing
def audits_all(
    before: Annotated[Optional[str], typer.Option(help="YYYY-DD-MMTHH:MM:SSZ")] = None,
    after: Annotated[Optional[str], typer.Option(help="YYYY-DD-MMTHH:MM:SSZ")] = None,
    sort_order: Annotated[Optional[SortOrder], typer.Option()] = None,
    sort_by: Annotated[Optional[SortBy], typer.Option()] = None,
    tenant: Optional[str] = None,
    region: Optional[str] = None,
):
    """Get all audits."""
    service = get_service(environment=region, tenant_id=tenant)

    results = service.audits.query.all_audits(
        AllAuditsInput(
            offset=0,
            limit=10000,
            before=before,
            after=after,
            sort_order=sort_order,
            sort_by=sort_by,
        )
    )

    normalized_results = TaegisAuditResultNormalizer(
        raw_results=results,
        service="audits",
        tenant_id=service.tenant_id,
        region=service.environment,
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


@app.command(name="search")
@tracing
def audits_search(
    audit_id: Annotated[Optional[str], typer.Option()] = None,
    log_type: Annotated[Optional[str], typer.Option()] = None,
    application: Annotated[Optional[str], typer.Option()] = None,
    request_type: Annotated[Optional[str], typer.Option()] = None,
    username: Annotated[Optional[str], typer.Option()] = None,
    email: Annotated[Optional[str], typer.Option()] = None,
    source: Annotated[Optional[str], typer.Option()] = None,
    target_rn: Annotated[Optional[str], typer.Option()] = None,
    action: Annotated[Optional[str], typer.Option()] = None,
    event_name: Annotated[Optional[str], typer.Option()] = None,
    event_desc: Annotated[Optional[str], typer.Option()] = None,
    trace_id: Annotated[Optional[str], typer.Option()] = None,
    url: Annotated[Optional[str], typer.Option()] = None,
    search_by_all: Annotated[Optional[str], typer.Option()] = None,
    before: Annotated[Optional[str], typer.Option(help="YYYY-DD-MMTHH:MM:SSZ")] = None,
    after: Annotated[Optional[str], typer.Option(help="YYYY-DD-MMTHH:MM:SSZ")] = None,
    sort_order: Annotated[Optional[SortOrder], typer.Option()] = None,
    sort_by: Annotated[Optional[SortBy], typer.Option()] = None,
    tenant: Annotated[Optional[str], typer.Option()] = None,
    region: Annotated[Optional[str], typer.Option()] = None,
):
    """Search for audits."""
    service = get_service(environment=region, tenant_id=tenant)

    results = service.audits.query.search_audits(
        AuditSearchInput(
            offset=0,
            limit=10000,
            id=audit_id,
            log_type=log_type,
            application=application,
            request_type=request_type,
            username=username,
            email=email,
            source=source,
            target_rn=target_rn,
            action=action,
            event_name=event_name,
            event_desc=event_desc,
            trace_id=trace_id,
            url=url,
            search_by_all=search_by_all,
            before=before,
            after=after,
            sort_by=sort_by,
            sort_order=sort_order,
        )
    )

    normalized_results = TaegisAuditResultNormalizer(
        raw_results=results,
        service="audits",
        tenant_id=service.tenant_id,
        region=service.environment,
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


@app.command(name="application-events")
@tracing
def audits_application_events(
    event_type: Annotated[AuditEventEnum, typer.Option()],
    tenant: Annotated[Optional[str], typer.Option()] = None,
    region: Annotated[Optional[str], typer.Option()] = None,
):
    """Get event name list."""
    service = get_service(environment=region, tenant_id=tenant)

    results = service.audits.query.list_application_events(application=event_type)

    normalized_results = TaegisAuditEventResultNormalizer(
        raw_results=results,
        service="audits",
        tenant_id=service.tenant_id,
        region=service.environment,
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results



"""Taegis Base Normalizer."""

from dataclasses import dataclass, field, asdict
from typing import Any, Dict, List, Union

import jinja2
import pandas as pd
from dataclasses_json import dataclass_json


@dataclass_json
@dataclass
class TaegisResultsNormalizer:
    """Taegis Results normalizer."""

    service: str
    tenant_id: str
    region: str
    raw_results: List[Dict[str, Any]] = field(default_factory=list)
    arguments: Dict[str, Any] = field(default_factory=dict)

    @property
    def normalizer(self) -> str:
        """Normalizer class name."""
        return str(self.__class__.__name__)

    @property
    def results(self) -> List[Dict[str, Any]]:
        """Results."""
        return self.raw_results

    @property
    def total_results(self) -> int:
        """Total Results."""
        return len(self.results)

    def _repr_markdown_(self):
        """Represent as markdown."""
        return self._display_template("taegis_results.md.jinja")

    def _display_template(self, template_name):  # pragma: no cover
        """Setup Jinja templating for markdown representation."""
        # template_name = "xdr_search_results.md.jinja"
        jinja_env = jinja2.Environment(
            loader=jinja2.PackageLoader("taegis_magic", "templates"),
            autoescape=jinja2.select_autoescape(["html", "xml"]),
        )

        def validate_int(value: int) -> Union[int, str]:
            """Validate normalizer integer values.

            Parameters
            ----------
            value : int
                Value to validate

            Returns
            -------
            Union[int, str]
                Return value or N/A
            """
            return value if value >= 0 else "N/A"

        jinja_env.filters["validate_int"] = validate_int

        template = jinja_env.get_template(template_name)
        return template.render(obj=self)


class TaegisResult(TaegisResultsNormalizer):
    """Generic single result normalizer."""

    raw_results: Any = field(default=None)

    @property
    def results(self):
        return [asdict(self.raw_results)]


class TaegisResults(TaegisResultsNormalizer):
    """Generic multiple results normalizer."""

    raw_results: List[Any] = field(default_factory=list)

    @property
    def results(self):
        return [asdict(r) for r in self.raw_results]


@dataclass_json
@dataclass
class DataFrameNormalizer(TaegisResultsNormalizer):
    raw_results: pd.DataFrame

    @property
    def results(self):
        return self.raw_results.to_dict(orient="records")

    def _repr_markdown_(self):
        """Represent as markdown."""
        return self.raw_results.to_html(index=False)

    def _repr_html_(self):
        """Represent as HTML."""
        return self.raw_results.to_html(index=False)
